import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "linux":
        assert False, "This is only available on Linux"


import asyncio
import contextlib
import logging
from typing import AsyncGenerator

from bumble import hci
from bumble.controller import Controller
from bumble.link import LocalLink
from bumble.transport import open_transport
from bumble.transport.common import Transport
from dbus_fast import BusType, Message, MessageType, Variant
from dbus_fast.aio.message_bus import MessageBus

from bleak._compat import timeout as async_timeout
from bleak.backends.bluezdbus import defs
from bleak.backends.bluezdbus.signals import MatchRules, add_match
from bleak.backends.bluezdbus.utils import assert_reply, get_dbus_authenticator

BLEAK_TEST_MANUFACTURER_ID = 0xB1EA

logger = logging.getLogger(__name__)


@contextlib.asynccontextmanager
async def open_message_bus() -> AsyncGenerator[MessageBus, None]:
    """
    Open a D-Bus message bus connection.
    """
    bus = MessageBus(bus_type=BusType.SYSTEM, auth=get_dbus_authenticator())
    try:
        await bus.connect()
        yield bus
    finally:
        bus.disconnect()
        await bus.wait_for_disconnect()


async def power_on_controller(
    bus: MessageBus, adapter_path: str, timeout: float = 2.0
) -> None:
    """
    Power on a Bluetooth controller via D-Bus.

    It may take some time for the adapter to fully configure itself after
    powering on. So we make multiple attempts until timeout is reached.
    """
    async with async_timeout(timeout):
        while True:
            try:
                reply = await bus.call(
                    Message(
                        destination=defs.BLUEZ_SERVICE,
                        path=adapter_path,
                        interface=defs.PROPERTIES_INTERFACE,
                        member="Set",
                        signature="ssv",
                        body=[defs.ADAPTER_INTERFACE, "Powered", Variant("b", True)],
                    )
                )
                assert_reply(reply)
                return
            except Exception as e:
                logger.debug("Failed to power on adapter at %s: %s", adapter_path, e)
                await asyncio.sleep(0.1)


@contextlib.asynccontextmanager
async def wait_for_new_adapter() -> (
    AsyncGenerator[tuple[MessageBus, asyncio.Future[str]], None]
):
    """
    Connect to D-Bus and wait for a new Bluetooth adapter to be added.

    Yields a Future that will be resolved with the adapter path when
    a new adapter is detected via InterfacesAdded signal.
    """
    async with open_message_bus() as bus:
        loop = asyncio.get_running_loop()
        adapter_path_future: asyncio.Future[str] = loop.create_future()

        def _on_interfaces_added(message: Message):
            if message.message_type != MessageType.SIGNAL:
                return
            if message.member != "InterfacesAdded":
                return

            obj_path, ifaces = message.body
            adapter = ifaces.get(defs.ADAPTER_INTERFACE)
            if not adapter:
                return

            # Check for our test manufacturer ID to identify the adapter
            if adapter["Manufacturer"].value != BLEAK_TEST_MANUFACTURER_ID:
                return

            if not adapter_path_future.done():
                adapter_path_future.set_result(obj_path)

        bus.add_message_handler(_on_interfaces_added)

        try:
            # Subscribe to InterfacesAdded signals
            reply = await add_match(
                bus,
                MatchRules(
                    interface=defs.OBJECT_MANAGER_INTERFACE,
                    member="InterfacesAdded",
                    arg0path="/org/bluez/",
                ),
            )
            assert_reply(reply)

            # Yield the future and bus for the caller to use
            yield bus, adapter_path_future
        finally:
            bus.remove_message_handler(_on_interfaces_added)


@contextlib.asynccontextmanager
async def open_bluez_bluetooth_controller_link(
    hci_transport_name: str,
) -> AsyncGenerator[LocalLink, None]:
    """
    Open a local link (virtual RF connection) to a bumble Bluetooth
    controller that is connected to BlueZ.
    """
    async with wait_for_new_adapter() as (bus, adapter_path_future):
        # Open a HCI transport connected to the OS
        async with await open_transport(hci_transport_name) as hci_transport:
            # Local link to create virtual RF connection between multiple Bluetooth Controllers
            link = LocalLink()

            # Bluetooth controller that BlueZ can connect to.
            # (This will register itself to the link.)
            bluez_controller = Controller(
                "BLEAK-TEST-BLUEZ",
                host_source=hci_transport.source,
                host_sink=hci_transport.sink,
                link=link,
            )
            bluez_controller.manufacturer_company_identifier = (
                BLEAK_TEST_MANUFACTURER_ID
            )

            # HACK: Work around Bumble missing feature combined with Linux kernel
            # requirement. https://github.com/google/bumble/issues/841
            #
            # According to the Bluetooth spec:
            #
            #   C24: [HCI_LE_Enhanced_Connection_Complete event is] Mandatory if
            #   the LE Controller supports Connection State and either LE Feature (LL
            #   Privacy) or LE Feature (Extended Advertising) is supported, otherwise optional if
            #   the LE Controller supports Connection State, otherwise excluded.
            #
            # And the Linux kernel enforces this in hci_le_create_conn_sync().
            # It will get a timeout if one of these features is enabled and the
            # Enhanced Connection Complete event is not sent.
            #
            # However, Bumble (as of 0.0.220) always sends HCI_LE_Connection_Complete_Event
            # in response to HCI_LE_Create_Connection_Command even when it should
            # be sending HCI_LE_Enhanced_Connection_Complete_Event.
            #
            # For now, we can work around the issue by disabling LL Privacy and
            # Extended Advertising features in the BlueZ controller.
            #
            # Ideally, this should be fixed in Bumble.
            bluez_controller.le_features &= ~(
                hci.LeFeatureMask.LL_PRIVACY | hci.LeFeatureMask.LE_EXTENDED_ADVERTISING
            )

            # Wait up to 5 seconds for the new adapter to appear via InterfacesAdded
            adapter_path = await asyncio.wait_for(adapter_path_future, timeout=5.0)
            logger.info(f"New adapter detected at {adapter_path}")

            # Ensure controller is powered on. This also ensures that BlueZ has fully
            # initialized the adapter and it is ready for use.
            await power_on_controller(bus, adapter_path)

            # We dont need the bus anymore. We have done everything needed with it.
            # Bleak will open its own D-Bus connection.
            bus.disconnect()
            await bus.wait_for_disconnect()

            # Yield the local link for use in tests
            yield link


@contextlib.asynccontextmanager
async def open_transport_with_bluez_vhci() -> AsyncGenerator[Transport, None]:
    """
    Create a bumble HCI Transport connected to BlueZ via the vhci driver and connect
    a Bluetooth controller for a peripheral device to it.
    """
    async with open_bluez_bluetooth_controller_link("vhci") as local_link:
        peripheral_controller = Controller("BLEAK-TEST-PERIPHERAL", link=local_link)
        yield Transport(peripheral_controller, peripheral_controller)



import pytest


def pytest_addoption(
    parser: pytest.Parser,
) -> None:
    parser.addoption(
        "--bleak-hci-transport",
        action="store",
        default=None,
        help="Bumble HCI transport moniker",
    )

    parser.addoption(
        "--bleak-bluez-vhci",
        action="store_true",
        default=False,
        help="Enable BlueZ VHCI Bumble HCI transport",
    )



import sys
from typing import Any, NewType, Optional, Protocol, TypeVar

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from Foundation import (
    NSUUID,
    NSArray,
    NSData,
    NSDictionary,
    NSError,
    NSErrorDomain,
    NSNumber,
    NSObject,
    NSString,
)
from libdispatch import dispatch_queue_t

CBErrorDomain: NSErrorDomain
CBATTErrorDomain: NSErrorDomain

class CBManager(NSObject):
    def state(self) -> CBManagerState: ...
    def authorization(self) -> CBManagerAuthorization: ...

TCBCentralManager = TypeVar("TCBCentralManager", bound=CBCentralManager)

class CBCentralManager(CBManager):
    @classmethod
    def initWithDelegate_queue_(
        cls,
        delegate: CBCentralManagerDelegate,
        queue: dispatch_queue_t,
    ) -> Self: ...
    @classmethod
    def initWithDelegate_queue_options_(
        cls,
        delegate: CBCentralManagerDelegate,
        queue: dispatch_queue_t,
        options: NSDictionary[str, Any],
    ) -> Optional[Self]: ...
    def connectPeripheral_options_(
        self,
        peripheral: CBPeripheral,
        options: Optional[NSDictionary[str, Any]],
    ) -> None: ...
    def cancelPeripheralConnection_(self, peripheral: CBPeripheral) -> None: ...
    def retrieveConnectedPeripheralsWithServices_(
        self, serviceUUIDs: NSArray[CBUUID]
    ) -> NSArray[CBPeripheral]: ...
    def retrievePeripheralsWithIdentifiers_(
        self, identifiers: NSArray[NSUUID]
    ) -> NSArray[CBPeripheral]: ...
    def scanForPeripheralsWithServices_options_(
        self,
        serviceUUIDs: Optional[NSArray[CBUUID]],
        options: Optional[NSDictionary[str, Any]],
    ) -> None: ...
    def stopScan(self) -> None: ...
    def isScanning(self) -> bool: ...
    @classmethod
    def supportsFeatures(cls, features: CBCentralManagerFeature) -> bool: ...
    def delegate(self) -> Optional[CBCentralManagerDelegate]: ...
    def registerForConnectionEventsWithOptions_(
        self, options: NSDictionary[str, Any]
    ) -> None: ...

CBConnectPeripheralOptionNotifyOnConnectionKey: str
CBConnectPeripheralOptionNotifyOnDisconnectionKey: str
CBConnectPeripheralOptionNotifyOnNotificationKey: str
CBConnectPeripheralOptionEnableTransportBridgingKey: str
CBConnectPeripheralOptionRequiresANCS: str
CBConnectPeripheralOptionStartDelayKey: str

CBCentralManagerScanOptionAllowDuplicatesKey: str
CBCentralManagerScanOptionSolicitedServiceUUIDsKey: str

CBCentralManagerFeature = NewType("CBCentralManagerFeature", int)

CBCentralManagerFeatureExtendedScanAndConnect: CBCentralManagerFeature

CBManagerState = NewType("CBManagerState", int)

CBManagerStatePoweredOff: CBManagerState
CBManagerStatePoweredOn: CBManagerState
CBManagerStateResetting: CBManagerState
CBManagerStateUnauthorized: CBManagerState
CBManagerStateUnknown: CBManagerState
CBManagerStateUnsupported: CBManagerState

CBManagerAuthorization = NewType("CBManagerAuthorization", int)

CBManagerAuthorizationAllowedAlways: CBManagerAuthorization
CBManagerAuthorizationDenied: CBManagerAuthorization
CBManagerAuthorizationNotDetermined: CBManagerAuthorization
CBManagerAuthorizationRestricted: CBManagerAuthorization

CBConnectionEvent = NewType("CBConnectionEvent", int)

CBConnectionEventPeerConnected: CBConnectionEvent
CBConnectionEventPeerDisconnected: CBConnectionEvent

class CBConnectionEventMatchingOption(str): ...

CBConnectionEventMatchingOptionPeripheralUUIDs: CBConnectionEventMatchingOption
CBConnectionEventMatchingOptionServiceUUIDs: CBConnectionEventMatchingOption

class CBCentralManagerDelegate(Protocol): ...

class CBPeer(NSObject):
    def identifier(self) -> NSUUID: ...

class CBPeripheral(CBPeer):
    def name(self) -> NSString: ...
    def delegate(self) -> CBPeripheralDelegate: ...
    def setDelegate_(self, delegate: CBPeripheralDelegate) -> None: ...
    def discoverServices_(self, serviceUUIDs: Optional[NSArray[CBUUID]]) -> None: ...
    def discoverIncludedServices_forService_(
        self, includedServiceUUIDs: NSArray[CBService], service: CBService
    ) -> None: ...
    def services(self) -> NSArray[CBService]: ...
    def discoverCharacteristics_forService_(
        self, characteristicUUIDs: Optional[NSArray[CBUUID]], service: CBService
    ) -> None: ...
    def discoverDescriptorsForCharacteristic_(
        self, characteristic: CBCharacteristic
    ) -> None: ...
    def readValueForCharacteristic_(self, characteristic: CBCharacteristic) -> None: ...
    def readValueForDescriptor_(self, descriptor: CBDescriptor) -> None: ...
    def writeValue_forCharacteristic_type_(
        self,
        data: NSData,
        characteristic: CBCharacteristic,
        type: CBCharacteristicWriteType,
    ) -> None: ...
    def writeValue_forDescriptor_(
        self, data: NSData, descriptor: CBDescriptor
    ) -> None: ...
    def maximumWriteValueLengthForType_(
        self, type: CBCharacteristicWriteType
    ) -> int: ...
    def setNotifyValue_forCharacteristic_(
        self, enabled: bool, characteristic: CBCharacteristic
    ) -> None: ...
    def state(self) -> CBPeripheralState: ...
    def canSendWriteWithoutResponse(self) -> bool: ...
    def readRSSI(self) -> None: ...
    def RSSI(self) -> NSNumber: ...

CBCharacteristicWriteType = NewType("CBCharacteristicWriteType", int)

CBCharacteristicWriteWithResponse: CBCharacteristicWriteType
CBCharacteristicWriteWithoutResponse: CBCharacteristicWriteType

CBPeripheralState = NewType("CBPeripheralState", int)

CBPeripheralStateDisconnected: CBPeripheralState
CBPeripheralStateConnecting: CBPeripheralState
CBPeripheralStateConnected: CBPeripheralState
CBPeripheralStateDisconnecting: CBPeripheralState

class CBPeripheralDelegate(Protocol):
    def peripheral_didDiscoverServices_(
        self, peripheral: CBPeripheral, error: Optional[NSError]
    ) -> None: ...
    def peripheral_didDiscoverIncludedServicesForService_error_(
        self, peripheral: CBPeripheral, service: CBService, error: Optional[NSError]
    ) -> None: ...
    def peripheral_didDiscoverCharacteristicsForService_error_(
        self, peripheral: CBPeripheral, service: CBService, error: Optional[NSError]
    ) -> None: ...
    def peripheral_didDiscoverDescriptorsForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None: ...
    def peripheral_didUpdateValueForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None: ...
    def peripheral_didUpdateValueForDescriptor_error_(
        self,
        peripheral: CBPeripheral,
        descriptor: CBDescriptor,
        error: Optional[NSError],
    ) -> None: ...
    def peripheral_didWriteValueForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None: ...
    def peripheral_didWriteValueForDescriptor_error_(
        self,
        peripheral: CBPeripheral,
        descriptor: CBDescriptor,
        error: Optional[NSError],
    ) -> None: ...
    def peripheralIsReadyToSendWriteWithoutResponse_(
        self, peripheral: CBPeripheral
    ) -> None: ...
    def peripheral_didUpdateNotificationStateForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None: ...
    def peripheral_didReadRSSI_error_(
        self,
        peripheral: CBPeripheral,
        rssi: NSNumber,
        error: Optional[NSError],
    ) -> None: ...
    def peripheralDidUpdateName_(self, peripheral: CBPeripheral) -> None: ...
    def peripheral_didModifyServices_(
        self, peripheral: CBPeripheral, invalidatedServices: NSArray[CBService]
    ) -> None: ...

class CBAttribute(NSObject):
    def UUID(self) -> CBUUID: ...

class CBService(CBAttribute):
    def peripheral(self) -> CBPeripheral: ...
    def isPrimary(self) -> bool: ...
    def characteristics(self) -> NSArray[CBCharacteristic]: ...
    def includedServices(self) -> Optional[NSArray[CBService]]: ...
    # Undocumented property
    def startHandle(self) -> int: ...

class CBUUID(NSObject):
    @classmethod
    def UUIDWithString_(cls, theString: str) -> CBUUID: ...
    @classmethod
    def UUIDWithData_(cls, theData: NSData) -> CBUUID: ...
    @classmethod
    def UUIDWithNSUUID_(cls, theUUID: NSUUID) -> CBUUID: ...
    def data(self) -> NSData: ...
    def UUIDString(self) -> NSString: ...

class CBCharacteristic(CBAttribute):
    def service(self) -> CBService: ...
    def value(self) -> Optional[NSData]: ...
    def descriptors(self) -> NSArray[CBDescriptor]: ...
    def properties(self) -> CBCharacteristicProperties: ...
    def isNotifying(self) -> bool: ...
    # Undocumented property
    def handle(self) -> int: ...

CBCharacteristicProperties = NewType("CBCharacteristicProperties", int)

CBCharacteristicPropertyBroadcast: CBCharacteristicProperties
CBCharacteristicPropertyRead: CBCharacteristicProperties
CBCharacteristicPropertyWriteWithoutResponse: CBCharacteristicProperties
CBCharacteristicPropertyWrite: CBCharacteristicProperties
CBCharacteristicPropertyNotify: CBCharacteristicProperties
CBCharacteristicPropertyIndicate: CBCharacteristicProperties
CBCharacteristicPropertyAuthenticatedSignedWrites: CBCharacteristicProperties
CBCharacteristicPropertyExtendedProperties: CBCharacteristicProperties
CBCharacteristicPropertyNotifyEncryptionRequired: CBCharacteristicProperties
CBCharacteristicPropertyIndicateEncryptionRequired: CBCharacteristicProperties

class CBDescriptor(CBAttribute):
    def characteristic(self) -> CBCharacteristic: ...
    def value(self) -> Optional[Any]: ...
    # Undocumented property
    def handle(self) -> int: ...

CBUUIDCharacteristicExtendedPropertiesString: NSString
CBUUIDCharacteristicUserDescriptionString: NSString
CBUUIDClientCharacteristicConfigurationString: NSString
CBUUIDServerCharacteristicConfigurationString: NSString
CBUUIDCharacteristicFormatString: NSString
CBUUIDCharacteristicAggregateFormatString: NSString
CBUUIDL2CAPPSMCharacteristicString: NSString



from typing import Literal, TypedDict

from bleak.assigned_numbers import CharacteristicPropertyName

# DBus Interfaces
OBJECT_MANAGER_INTERFACE = "org.freedesktop.DBus.ObjectManager"
PROPERTIES_INTERFACE = "org.freedesktop.DBus.Properties"

# Bluez specific DBUS
BLUEZ_SERVICE = "org.bluez"
ADAPTER_INTERFACE = "org.bluez.Adapter1"
ADVERTISEMENT_MONITOR_INTERFACE = "org.bluez.AdvertisementMonitor1"
ADVERTISEMENT_MONITOR_MANAGER_INTERFACE = "org.bluez.AdvertisementMonitorManager1"
DEVICE_INTERFACE = "org.bluez.Device1"
BATTERY_INTERFACE = "org.bluez.Battery1"

# GATT interfaces
GATT_MANAGER_INTERFACE = "org.bluez.GattManager1"
GATT_PROFILE_INTERFACE = "org.bluez.GattProfile1"
GATT_SERVICE_INTERFACE = "org.bluez.GattService1"
GATT_CHARACTERISTIC_INTERFACE = "org.bluez.GattCharacteristic1"
GATT_DESCRIPTOR_INTERFACE = "org.bluez.GattDescriptor1"

# BlueZ error names
BLUEZ_ERROR_DOES_NOT_EXIST = "org.bluez.Error.DoesNotExist"
BLUEZ_ERROR_FAILED = "org.bluez.Error.Failed"
BLUEZ_ERROR_IMPROPERLY_CONFIGURED = "org.bluez.Error.ImproperlyConfigured"
BLUEZ_ERROR_IN_PROGRESS = "org.bluez.Error.InProgress"
BLUEZ_ERROR_INVALID_ARGUMENT = "org.bluez.Error.InvalidArguments"
BLUEZ_ERROR_INVALID_OFFSET = "org.bluez.Error.InvalidOffset"
BLUEZ_ERROR_INVALID_VALUE_LENGTH = "org.bluez.Error.InvalidValueLength"
BLUEZ_ERROR_NOT_AUTHORIZED = "org.bluez.Error.NotAuthorized"
BLUEZ_ERROR_NOT_PERMITTED = "org.bluez.Error.NotPermitted"
BLUEZ_ERROR_NOT_READY = "org.bluez.Error.NotReady"
BLUEZ_ERROR_NOT_SUPPORTED = "org.bluez.Error.NotSupported"

# D-Bus properties for interfaces
# https://github.com/bluez/bluez/blob/master/doc/org.bluez.Adapter.rst


class Adapter1(TypedDict):
    Address: str
    Name: str
    Alias: str
    Class: int
    Powered: bool
    Discoverable: bool
    Pairable: bool
    PairableTimeout: int
    DiscoverableTimeout: int
    Discovering: int
    UUIDs: list[str]
    Modalias: str
    Roles: list[str]
    ExperimentalFeatures: list[str]


# https://github.com/bluez/bluez/blob/master/doc/org.bluez.AdvertisementMonitor.rst


class AdvertisementMonitor1(TypedDict):
    Type: str
    RSSILowThreshold: int
    RSSIHighThreshold: int
    RSSILowTimeout: int
    RSSIHighTimeout: int
    RSSISamplingPeriod: int
    Patterns: list[tuple[int, int, bytes]]


# https://github.com/bluez/bluez/blob/master/doc/org.bluez.AdvertisementMonitorManager.rst


class AdvertisementMonitorManager1(TypedDict):
    SupportedMonitorTypes: list[str]
    SupportedFeatures: list[str]


# https://github.com/bluez/bluez/blob/master/doc/org.bluez.Battery.rst


class Battery1(TypedDict):
    SupportedMonitorTypes: list[str]
    SupportedFeatures: list[str]


# https://github.com/bluez/bluez/blob/master/doc/org.bluez.Device.rst


class Device1(TypedDict):
    Address: str
    AddressType: str
    Name: str
    Icon: str
    Class: int
    Appearance: int
    UUIDs: list[str]
    Paired: bool
    Bonded: bool
    Connected: bool
    Trusted: bool
    Blocked: bool
    WakeAllowed: bool
    Alias: str
    Adapter: str
    LegacyPairing: bool
    Modalias: str
    RSSI: int
    TxPower: int
    ManufacturerData: dict[int, bytes]
    ServiceData: dict[str, bytes]
    ServicesResolved: bool
    AdvertisingFlags: bytes
    AdvertisingData: dict[int, bytes]


# https://github.com/bluez/bluez/blob/master/doc/org.bluez.GattService.rst


class GattService1(TypedDict):
    UUID: str
    Primary: bool
    Device: str
    Includes: list[str]
    # Handle is server-only and not available in Bleak


class GattCharacteristic1(TypedDict):
    UUID: str
    Service: str
    Value: bytes
    WriteAcquired: bool
    NotifyAcquired: bool
    Notifying: bool
    Flags: list[CharacteristicPropertyName]
    # "MTU" property was added in BlueZ 5.62.
    # It may missing when operating with an older stack.
    MTU: int
    # Handle is server-only and not available in Bleak


class GattDescriptor1(TypedDict):
    UUID: str
    Characteristic: str
    Value: bytes
    Flags: list[
        Literal[
            "read",
            "write",
            "encrypt-read",
            "encrypt-write",
            "encrypt-authenticated-read",
            "encrypt-authenticated-write",
            # "secure-read" and "secure-write" are server-only and not available in Bleak
            "authorize",
        ]
    ]
    # Handle is server-only and not available in Bleak



import asyncio

import pytest
from bumble.device import Device

from bleak import BleakClient
from bleak._compat import timeout as async_timeout
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    find_ble_device,
)


async def test_connect(bumble_peripheral: Device):
    """Connecting to a BLE device is possible."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device) as client:
        assert client.name == bumble_peripheral.name


async def test_connect_multiple_times(bumble_peripheral: Device):
    """Connecting to a BLE device multiple times is possible."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device):
        pass

    await bumble_peripheral.start_advertising()

    async with BleakClient(device):
        pass


async def test_connect_timeout(bumble_peripheral: Device):
    """Connecting to a removed BLE device times out."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    await bumble_peripheral.stop_advertising()

    with pytest.raises(asyncio.TimeoutError):
        async with BleakClient(device, timeout=1.0):
            pass


async def test_is_connected(bumble_peripheral: Device):
    """Check if a connection is connected is working."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    client = BleakClient(device)

    assert client.is_connected is False
    async with BleakClient(device) as client:
        assert client.is_connected is True
    assert client.is_connected is False


async def test_disconnect_callback(bumble_peripheral: Device):
    """Check if disconnect callback is called."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await find_ble_device(bumble_peripheral)

    disconnected_client_future: asyncio.Future[BleakClient] = asyncio.Future()

    def disconnected_callback(client: BleakClient):
        disconnected_client_future.set_result(client)

    async with BleakClient(device, disconnected_callback) as client:
        # Disconnect from virtual device side
        virtual_connection = list(bumble_peripheral.connections.values())[0]
        await virtual_connection.disconnect()

        # Wait for disconnected callback to be called
        async with async_timeout(5):
            disconnected_client = await disconnected_client_future
        assert disconnected_client is client



import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    if sys.platform != "win32":
        assert False, "This backend is only available on Windows"

import asyncio
import ctypes
from ctypes import wintypes
from enum import IntEnum

from bleak._compat import timeout as async_timeout
from bleak.exc import BleakError


def _check_result(result: int, func: Any, args: Any) -> Any:
    if not result:
        raise ctypes.WinError()

    return args


def check_hresult(result: int, func: Any, args: Any) -> Any:
    if result:
        raise ctypes.WinError(result)

    return args


# not defined in wintypes
_UINT_PTR = wintypes.WPARAM

# https://learn.microsoft.com/en-us/windows/win32/api/winuser/nc-winuser-timerproc
_TIMERPROC = ctypes.WINFUNCTYPE(
    None, wintypes.HWND, _UINT_PTR, wintypes.UINT, wintypes.DWORD
)

# https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-settimer
_SET_TIMER_PROTOTYPE = ctypes.WINFUNCTYPE(
    _UINT_PTR, wintypes.HWND, _UINT_PTR, wintypes.UINT, _TIMERPROC
)
_SET_TIMER_PARAM_FLAGS = (
    (1, "hwnd", None),
    (1, "nidevent"),
    (1, "uelapse"),
    (1, "lptimerfunc", None),
)
_SetTimer = _SET_TIMER_PROTOTYPE(
    ("SetTimer", ctypes.windll.user32), _SET_TIMER_PARAM_FLAGS
)
_SetTimer.errcheck = _check_result  # type: ignore[assignment]

# https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-killtimer
_KILL_TIMER_PROTOTYPE = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, _UINT_PTR)
_KILL_TIMER_PARAM_FLAGS = (
    (1, "hwnd", None),
    (1, "uidevent"),
)
_KillTimer = _KILL_TIMER_PROTOTYPE(
    ("KillTimer", ctypes.windll.user32), _KILL_TIMER_PARAM_FLAGS
)

# https://learn.microsoft.com/en-us/windows/win32/api/combaseapi/nf-combaseapi-cogetapartmenttype
_CO_GET_APARTMENT_TYPE_PROTOTYPE = ctypes.WINFUNCTYPE(
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
)
_CO_GET_APARTMENT_TYPE_PARAM_FLAGS = (
    (1, "papttype", None),
    (1, "paptqualifier", None),
)
_CoGetApartmentType = _CO_GET_APARTMENT_TYPE_PROTOTYPE(
    ("CoGetApartmentType", ctypes.windll.ole32), _CO_GET_APARTMENT_TYPE_PARAM_FLAGS
)
_CoGetApartmentType.errcheck = check_hresult  # type: ignore[assignment]

_CO_E_NOTINITIALIZED = -2147221008


# https://learn.microsoft.com/en-us/windows/win32/api/objidl/ne-objidl-apttype
class _AptType(IntEnum):
    CURRENT = -1
    STA = 0
    MTA = 1
    NA = 2
    MAIN_STA = 3


# https://learn.microsoft.com/en-us/windows/win32/api/objidl/ne-objidl-apttypequalifier
class _AptQualifierType(IntEnum):
    NONE = 0
    IMPLICIT_MTA = 1
    NA_ON_MTA = 2
    NA_ON_STA = 3
    NA_ON_IMPLICIT_STA = 4
    NA_ON_MAIN_STA = 5
    APPLICATION_STA = 6
    RESERVED_1 = 7


def _get_apartment_type() -> tuple[_AptType, _AptQualifierType]:
    """
    Calls CoGetApartmentType to get the current apartment type and qualifier.

    Returns:
        The current apartment type and qualifier.
    Raises:
        OSError: If the call to CoGetApartmentType fails.
    """
    api_type = ctypes.c_int()
    api_type_qualifier = ctypes.c_int()
    _CoGetApartmentType(ctypes.byref(api_type), ctypes.byref(api_type_qualifier))
    return _AptType(api_type.value), _AptQualifierType(api_type_qualifier.value)


async def assert_mta() -> None:
    """
    Asserts that the current apartment type is MTA.

    Raises:
        BleakError:
            If the current apartment type is not MTA and there is no Windows
            message loop running.

    .. versionadded:: 0.22

    .. versionchanged:: 0.22.2

        Function is now async and will not raise if the current apartment type
        is STA and the Windows message loop is running.
    """
    if hasattr(allow_sta, "_allowed"):
        return

    try:
        apt_type, _ = _get_apartment_type()
    except OSError as e:
        # All is OK if not initialized yet. WinRT will initialize it.
        if e.winerror == _CO_E_NOTINITIALIZED:
            return

        raise

    if apt_type == _AptType.MTA:
        # if we get here, WinRT probably set the apartment type to MTA and all
        # is well, we don't need to check again
        setattr(allow_sta, "_allowed", True)
        return

    event = asyncio.Event()

    def wait_event(hwnd: int, uidevent: int, uelapse: int, dwtime: int) -> None:
        event.set()

    # have to keep a reference to the callback or it will be garbage collected
    # before it is called
    callback = _TIMERPROC(wait_event)

    # set a timer to see if we get a callback to ensure the windows event loop
    # is running
    timer = _SetTimer(None, 1, 0, callback)

    try:
        async with async_timeout(0.5):
            await event.wait()
    except asyncio.TimeoutError:
        raise BleakError(
            "Thread is configured for Windows GUI but callbacks are not working."
            + (
                " Suspect unwanted side effects from importing 'pythoncom'."
                if "pythoncom" in sys.modules
                else ""
            )
        )
    else:
        # if the windows event loop is running, we assume it is going to keep
        # running and we don't need to check again
        setattr(allow_sta, "_allowed", True)
    finally:
        _KillTimer(None, timer)


def allow_sta() -> None:
    """
    Suppress check for MTA thread type and allow STA.

    Bleak will hang forever if the current thread is not MTA - unless there is
    a Windows event loop running that is properly integrated with asyncio in
    Python.

    If your program meets that condition, you must call this function do disable
    the check for MTA. If your program doesn't have a graphical user interface
    you probably shouldn't call this function. and use ``uninitialize_sta()``
    instead.

    .. versionadded:: 0.22.1
    """
    setattr(allow_sta, "_allowed", True)


def uninitialize_sta():
    """
    Uninitialize the COM library on the current thread if it was not initialized
    as MTA.

    This is intended to undo the implicit initialization of the COM library as STA
    by packages like pywin32.

    It should be called as early as possible in your application after the
    offending package has been imported.

    .. versionadded:: 0.22
    """

    try:
        _get_apartment_type()
    except OSError as e:
        # All is OK if not initialized yet. WinRT will initialize it.
        if e.winerror == _CO_E_NOTINITIALIZED:
            return
    else:
        ctypes.windll.ole32.CoUninitialize()



import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "linux":
        assert False, "This backend is only available on Linux"

import logging
from collections.abc import Callable, Coroutine
from typing import Any, Literal, Optional
from warnings import warn

from dbus_fast import Variant

from bleak._compat import override
from bleak.args.bluez import BlueZDiscoveryFilters as _BlueZDiscoveryFilters
from bleak.args.bluez import BlueZScannerArgs as _BlueZScannerArgs
from bleak.backends.bluezdbus.defs import Device1
from bleak.backends.bluezdbus.manager import get_global_bluez_manager
from bleak.backends.scanner import (
    AdvertisementData,
    AdvertisementDataCallback,
    BaseBleakScanner,
)
from bleak.exc import BleakError

logger = logging.getLogger(__name__)


_DEPRECATED: dict[str, Any] = {
    "BlueZDiscoveryFilters": _BlueZDiscoveryFilters,
    "BlueZScannerArgs": _BlueZScannerArgs,
}


def __getattr__(name: str):
    if value := _DEPRECATED.get(name):
        warn(
            f"importing {name} from bleak.backends.bluezdbus.scanner is deprecated, use bleak.args.bluez instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return value
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class BleakScannerBlueZDBus(BaseBleakScanner):
    """The native Linux Bleak BLE Scanner.

    For possible values for `filters`, see the parameters to the
    ``SetDiscoveryFilter`` method in the `BlueZ docs
    <https://github.com/bluez/bluez/blob/master/doc/org.bluez.Adapter.rst#void-setdiscoveryfilterdict-filter>`_

    Args:
        detection_callback:
            Optional function that will be called each time a device is
            discovered or advertising data has changed.
        service_uuids:
            Optional list of service UUIDs to filter on. Only advertisements
            containing this advertising data will be received. Specifying this
            also enables scanning while the screen is off on Android.
        scanning_mode:
            Set to ``"passive"`` to avoid the ``"active"`` scanning mode.
        **bluez:
            Dictionary of arguments specific to the BlueZ backend.
    """

    def __init__(
        self,
        detection_callback: Optional[AdvertisementDataCallback],
        service_uuids: Optional[list[str]],
        scanning_mode: Literal["active", "passive"],
        *,
        bluez: _BlueZScannerArgs,
        **kwargs: Any,
    ):
        super().__init__(detection_callback, service_uuids)

        self._scanning_mode = scanning_mode
        self._adapter = bluez.get("adapter")

        # callback from manager for stopping scanning if it has been started
        self._stop: Optional[Callable[[], Coroutine[Any, Any, None]]] = None

        # Discovery filters

        self._filters: dict[str, Variant] = {}

        self._filters["Transport"] = Variant("s", "le")
        self._filters["DuplicateData"] = Variant("b", False)

        if self._service_uuids:
            self._filters["UUIDs"] = Variant("as", self._service_uuids)

        filters = bluez.get("filters")

        if filters is not None:
            self.set_scanning_filter(filters=filters)

        self._or_patterns = bluez.get("or_patterns")

        if self._scanning_mode == "passive" and service_uuids:
            logger.warning(
                "service uuid filtering is not implemented for passive scanning, use bluez or_patterns as a workaround"
            )

        if self._scanning_mode == "passive" and not self._or_patterns:
            raise BleakError("passive scanning mode requires bluez or_patterns")

    @override
    async def start(self) -> None:
        manager = await get_global_bluez_manager()

        if self._adapter:
            adapter_path = f"/org/bluez/{self._adapter}"
        else:
            adapter_path = manager.get_default_adapter()

        self.seen_devices = {}

        if self._scanning_mode == "passive":
            assert self._or_patterns is not None  # should be checked in __init__

            self._stop = await manager.passive_scan(
                adapter_path,
                self._or_patterns,
                self._handle_advertising_data,
                self._handle_device_removed,
            )
        else:
            self._stop = await manager.active_scan(
                adapter_path,
                self._filters,
                self._handle_advertising_data,
                self._handle_device_removed,
            )

    @override
    async def stop(self) -> None:
        if self._stop:
            # avoid reentrancy
            stop, self._stop = self._stop, None

            await stop()

    def set_scanning_filter(self, **kwargs: Any) -> None:
        """Sets OS level scanning filters for the BleakScanner.

        For possible values for `filters`, see the parameters to the
        ``SetDiscoveryFilter`` method in the `BlueZ docs
        <https://github.com/bluez/bluez/blob/master/doc/org.bluez.Adapter.rst#void-setdiscoveryfilterdict-filter>`_

        See variant types here: <https://python-dbus-next.readthedocs.io/en/latest/type-system/>

        Keyword Args:
            filters (dict): A dict of filters to be applied on discovery.

        """
        for k, v in kwargs.get("filters", {}).items():
            if k == "UUIDs":
                self._filters[k] = Variant("as", v)
            elif k == "RSSI":
                self._filters[k] = Variant("n", v)
            elif k == "Pathloss":
                self._filters[k] = Variant("n", v)
            elif k == "Transport":
                self._filters[k] = Variant("s", v)
            elif k == "DuplicateData":
                self._filters[k] = Variant("b", v)
            elif k == "Discoverable":
                self._filters[k] = Variant("b", v)
            elif k == "Pattern":
                self._filters[k] = Variant("s", v)
            else:
                logger.warning("Filter '%s' is not currently supported.", k)

    # Helper methods

    def _handle_advertising_data(self, path: str, props: Device1) -> None:
        """
        Handles advertising data received from the BlueZ manager instance.

        Args:
            path: The D-Bus object path of the device.
            props: The D-Bus object properties of the device.
        """
        _service_uuids = props.get("UUIDs", [])

        if not self.is_allowed_uuid(_service_uuids):
            return

        # Get all the information wanted to pack in the advertisement data
        _local_name = props.get("Name")
        _manufacturer_data = {
            k: bytes(v) for k, v in props.get("ManufacturerData", {}).items()
        }
        _service_data = {k: bytes(v) for k, v in props.get("ServiceData", {}).items()}

        # Get tx power data
        tx_power = props.get("TxPower")

        # Pack the advertisement data
        advertisement_data = AdvertisementData(
            local_name=_local_name,
            manufacturer_data=_manufacturer_data,
            service_data=_service_data,
            service_uuids=_service_uuids,
            tx_power=tx_power,
            rssi=props.get("RSSI", -127),
            platform_data=(path, props),
        )

        device = self.create_or_update_device(
            path,
            props["Address"],
            # BlueZ generates a name based on the address if no name is available.
            # To match other backends, we replace this with None.
            (
                None
                if props["Alias"] == props["Address"].replace(":", "-")
                else props["Alias"]
            ),
            {"path": path, "props": props},
            advertisement_data,
        )

        self.call_detection_callbacks(device, advertisement_data)

    def _handle_device_removed(self, device_path: str) -> None:
        """
        Handles a device being removed from BlueZ.
        """
        try:
            del self.seen_devices[device_path]
        except KeyError:
            # The device will not have been added to self.seen_devices if no
            # advertising data was received, so this is expected to happen
            # occasionally.
            pass



import asyncio
import contextlib
import sys

import pytest
from bumble import data_types
from bumble.core import UUID
from bumble.device import Device

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.uuids import normalize_uuid_str
from tests.integration.conftest import configure_and_power_on_bumble_peripheral

DEFAULT_TIMEOUT = 5.0


async def test_discover(bumble_peripheral: Device):
    """Scanner discovery is finding the device."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    devices = await BleakScanner.discover(return_adv=True)
    filtered_devices = list(
        filter(
            lambda device: device[1].local_name == bumble_peripheral.name,
            devices.values(),
        )
    )

    assert len(filtered_devices) == 1
    assert filtered_devices[0][1].local_name == bumble_peripheral.name
    if sys.platform != "darwin":
        # accessing the address is not possible on macOS
        assert filtered_devices[0][0].address == str(bumble_peripheral.static_address)


async def test_find_by_address(bumble_peripheral: Device):
    """Scanner is finding the device by address."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    device = await BleakScanner.find_device_by_address(
        bumble_peripheral.static_address.to_string(), cb={"use_bdaddr": True}
    )
    assert device is not None


@pytest.mark.parametrize("service_uuid_available", [True, False])
async def test_discover_filter_by_service_uuid(
    bumble_peripheral: Device,
    service_uuid_available: bool,
):
    """Scanner discovery is filtering service uuids correctly."""
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        additional_adv_data=(
            [data_types.IncompleteListOf16BitServiceUUIDs([UUID(0x180F)])]
            if service_uuid_available
            else []
        ),
    )

    found_adv_data_future: asyncio.Future[AdvertisementData] = asyncio.Future()

    def detection_callback(device: BLEDevice, adv_data: AdvertisementData):
        if device.name == bumble_peripheral.name:
            found_adv_data_future.set_result(adv_data)

    async with BleakScanner(
        detection_callback,
        service_uuids=[normalize_uuid_str("180f")],
    ):
        found_adv_data = None
        with contextlib.suppress(asyncio.TimeoutError):
            found_adv_data = await asyncio.wait_for(
                found_adv_data_future, timeout=DEFAULT_TIMEOUT
            )

    if service_uuid_available:
        assert found_adv_data is not None
    else:
        assert found_adv_data is None


async def test_adv_data_simple(bumble_peripheral: Device):
    """Simple advertising data is parsed correct."""
    await configure_and_power_on_bumble_peripheral(bumble_peripheral)

    found_adv_data_future: asyncio.Future[AdvertisementData] = asyncio.Future()

    def detection_callback(device: BLEDevice, adv_data: AdvertisementData):
        if device.name == bumble_peripheral.name:
            found_adv_data_future.set_result(adv_data)

    async with BleakScanner(detection_callback):
        found_adv_data = await asyncio.wait_for(
            found_adv_data_future, timeout=DEFAULT_TIMEOUT
        )

    assert found_adv_data is not None
    assert found_adv_data.local_name == bumble_peripheral.name
    assert found_adv_data.manufacturer_data == {}
    assert found_adv_data.service_data == {}
    assert found_adv_data.service_uuids == []
    assert found_adv_data.tx_power is None
    assert found_adv_data.platform_data

    # Verify that this value is an integer and not some other
    # type from a ffi binding framework.
    assert isinstance(found_adv_data.rssi, int)

    # The rssi can vary. So we only check for a plausible range.
    assert -127 <= found_adv_data.rssi <= 20


async def test_adv_data_complex(bumble_peripheral: Device):
    """Complex advertising data is parsed correct."""
    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        [
            data_types.ManufacturerSpecificData(0x1234, b"MFG"),
            data_types.IncompleteListOf16BitServiceUUIDs([UUID(0x180F)]),
            data_types.TxPowerLevel(123),
            data_types.ServiceData16BitUUID(UUID(0x180F), b"SER"),
        ],
    )

    found_adv_data_future: asyncio.Future[AdvertisementData] = asyncio.Future()

    def detection_callback(device: BLEDevice, adv_data: AdvertisementData):
        if device.name == bumble_peripheral.name:
            found_adv_data_future.set_result(adv_data)

    async with BleakScanner(detection_callback):
        found_adv_data = await asyncio.wait_for(
            found_adv_data_future, timeout=DEFAULT_TIMEOUT
        )

    assert found_adv_data is not None
    assert found_adv_data.local_name == bumble_peripheral.name
    assert found_adv_data.manufacturer_data == {0x1234: b"MFG"}
    assert found_adv_data.service_data == {
        "0000180f-0000-1000-8000-00805f9b34fb": b"SER"
    }
    assert found_adv_data.service_uuids == ["0000180f-0000-1000-8000-00805f9b34fb"]
    assert found_adv_data.tx_power == 123
    assert found_adv_data.platform_data

    # Verify that this value is an integer and not some other
    # type from a ffi binding framework.
    assert isinstance(found_adv_data.rssi, int)

    # The rssi can vary. So we only check for a plausible range.
    assert -127 <= found_adv_data.rssi <= 20



import asyncio
import logging
from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar

from bleak._compat import TypeVarTuple, Unpack

_Ts = TypeVarTuple("_Ts")


logger = logging.getLogger(__name__)


def try_call_soon_threadsafe(
    event_loop: asyncio.AbstractEventLoop,
    callback: Callable[[Unpack[_Ts]], Any],
    *args: Unpack[_Ts],
) -> None:
    """Call a callback on the event loop thread, handling closed loop errors gracefully."""
    try:
        event_loop.call_soon_threadsafe(callback, *args)
    except RuntimeError:
        # Likely caused by loop being closed
        logger.debug("unraisable exception", exc_info=True)  # pragma: no cover


_P = ParamSpec("_P")
_TReturn = TypeVar("_TReturn")


def external_thread_callback(func: Callable[_P, _TReturn]) -> Callable[_P, _TReturn]:
    """
    Decorator for callbacks invoked from external non-Python thread.

    This is a no-op placeholder that can be overridden in tests to enable
    coverage tracing for external threads.
    """
    return func



"""
Communicating with Bluetooth hardware requires calling OS-specific APIs. These
are abstracted as "backends" in Bleak.

The backend will be automatically selected based on the operating system Bleak
is running on. In some cases, this may also depend on a specific runtime, like
Pythonista on iOS.
"""

import enum
import os
import platform
import sys

from bleak.exc import BleakError


class BleakBackend(str, enum.Enum):
    """
    Identifiers for available built-in Bleak backends.

    .. versionadded:: 2.0
    """

    P4ANDROID = "p4android"
    """
    Python for Android backend.
    """

    BLUEZ_DBUS = "bluez_dbus"
    """
    BlueZ D-Bus backend for Linux.
    """

    PYTHONISTA_CB = "pythonista_cb"
    """
    Pythonista CoreBluetooth backend for iOS and macOS.
    """

    CORE_BLUETOOTH = "core_bluetooth"
    """
    CoreBluetooth backend for macOS.
    """

    WIN_RT = "win_rt"
    """
    Windows Runtime backend for Windows.
    """


def get_default_backend() -> BleakBackend:
    """
    Returns the preferred backend for the current platform/environment.

    .. versionadded:: 2.0
    """
    if os.environ.get("P4A_BOOTSTRAP") is not None:
        return BleakBackend.P4ANDROID

    if platform.system() == "Linux":
        return BleakBackend.BLUEZ_DBUS

    if sys.platform == "ios" and "Pythonista3.app" in sys.executable:
        # Must be resolved before checking for "Darwin" (macOS),
        # as both the Pythonista app for iOS and macOS
        # return "Darwin" from platform.system()
        return BleakBackend.PYTHONISTA_CB

    if platform.system() == "Darwin":
        return BleakBackend.CORE_BLUETOOTH

    if platform.system() == "Windows":
        return BleakBackend.WIN_RT

    raise BleakError(f"Unsupported platform: {platform.system()}")



"""
Service Explorer
----------------

An example showing how to access and print out the services, characteristics and
descriptors of a connected GATT server.

Created on 2019-03-25 by hbldh <henrik.blidh@nedomkull.com>

"""

import argparse
import asyncio
import logging
from typing import Optional

from bleak import BleakClient, BleakScanner

logger = logging.getLogger(__name__)


class Args(argparse.Namespace):
    name: Optional[str]
    address: Optional[str]
    macos_use_bdaddr: bool
    services: list[str]
    pair: bool
    debug: bool


async def main(args: Args):
    logger.info("starting scan...")

    if args.address:
        device = await BleakScanner.find_device_by_address(
            args.address, cb={"use_bdaddr": args.macos_use_bdaddr}
        )
        if device is None:
            logger.error("could not find device with address '%s'", args.address)
            return
    elif args.name:
        device = await BleakScanner.find_device_by_name(
            args.name, cb={"use_bdaddr": args.macos_use_bdaddr}
        )
        if device is None:
            logger.error("could not find device with name '%s'", args.name)
            return
    else:
        raise ValueError("Either --name or --address must be provided")

    logger.info("connecting to device...")

    async with BleakClient(
        device,
        pair=args.pair,
        services=args.services,
        # Give the user plenty of time to enter a PIN code if paring is required.
        timeout=90 if args.pair else 10,
    ) as client:
        logger.info("connected to %s (%s)", client.name, client.address)

        for service in client.services:
            logger.info("[Service] %s", service)

            for char in service.characteristics:
                if "read" in char.properties:
                    try:
                        value = await client.read_gatt_char(char)
                        extra = f", Value: {value}"
                    except Exception as e:
                        extra = f", Error: {e}"
                else:
                    extra = ""

                if "write-without-response" in char.properties:
                    extra += f", Max write w/o rsp size: {char.max_write_without_response_size}"

                logger.info(
                    "  [Characteristic] %s (%s)%s",
                    char,
                    ",".join(char.properties),
                    extra,
                )

                for descriptor in char.descriptors:
                    try:
                        value = await client.read_gatt_descriptor(descriptor)
                        logger.info("    [Descriptor] %s, Value: %r", descriptor, value)
                    except Exception as e:
                        logger.error("    [Descriptor] %s, Error: %s", descriptor, e)

        logger.info("disconnecting...")

    logger.info("disconnected")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    device_group = parser.add_mutually_exclusive_group(required=True)

    device_group.add_argument(
        "--name",
        metavar="<name>",
        help="the name of the bluetooth device to connect to",
    )
    device_group.add_argument(
        "--address",
        metavar="<address>",
        help="the address of the bluetooth device to connect to",
    )

    parser.add_argument(
        "--macos-use-bdaddr",
        action="store_true",
        help="when true use Bluetooth address instead of UUID on macOS",
    )

    parser.add_argument(
        "--services",
        nargs="+",
        metavar="<uuid>",
        help="if provided, only enumerate matching service(s)",
    )

    parser.add_argument(
        "--pair",
        action="store_true",
        help="pair with the device before connecting if not already paired",
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="sets the log level to debug",
    )

    args = parser.parse_args(namespace=Args())

    log_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(
        level=log_level,
        format="%(asctime)-15s %(name)-8s %(levelname)s: %(message)s",
    )

    asyncio.run(main(args))



import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

from typing import Any, Optional, TypeGuard, overload

from CoreBluetooth import (
    CBUUID,
    CBManagerState,
    CBManagerStatePoweredOff,
    CBManagerStatePoweredOn,
    CBManagerStateResetting,
    CBManagerStateUnauthorized,
    CBManagerStateUnknown,
    CBManagerStateUnsupported,
    CBUUIDCharacteristicExtendedPropertiesString,
    CBUUIDCharacteristicUserDescriptionString,
    CBUUIDClientCharacteristicConfigurationString,
    CBUUIDServerCharacteristicConfigurationString,
)
from Foundation import NSNumber, NSString

from bleak.uuids import normalize_uuid_str


def cb_uuid_to_str(uuid: CBUUID) -> str:
    """Converts a CoreBluetooth UUID to a Python string.

    If ``uuid`` is a 16-bit UUID, it is assumed to be a Bluetooth GATT UUID
    (``0000xxxx-0000-1000-8000-00805f9b34fb``).

    Args
        uuid: The UUID.

    Returns:
        The UUID as a lower case Python string (``xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxx``)
    """
    return normalize_uuid_str(uuid.UUIDString())


@overload
def to_optional_str(value: NSString) -> str: ...
@overload
def to_optional_str(value: None) -> None: ...


def to_optional_str(value: Optional[NSString]) -> Optional[str]:
    """Converts an NSString to a Python string or None.

    Args:
        value: The NSString or None.

    Returns:
        The Python string or None.
    """
    if value is None:
        return None

    return str(value)


@overload
def to_optional_int(value: NSNumber) -> int: ...
@overload
def to_optional_int(value: None) -> None: ...


def to_optional_int(value: Optional[NSNumber]) -> Optional[int]:
    """Converts an NSNumber to a Python int or None.

    Args:
        value: The NSNumber or None.

    Returns:
        The Python int or None.
    """
    if value is None:
        return None

    return int(value)


# Most descriptors are returned as NSData (raw bytes), but some of them
# are returned as NSNumber or NSString.
# See: https://developer.apple.com/documentation/corebluetooth/characteristic-descriptors
_DESCRIPTOR_TYPE_NSNUMBER = (
    normalize_uuid_str(CBUUIDCharacteristicExtendedPropertiesString),  # 0x2900
    normalize_uuid_str(CBUUIDClientCharacteristicConfigurationString),  # 0x2902
    normalize_uuid_str(CBUUIDServerCharacteristicConfigurationString),  # 0x2903
)
_DESCRIPTOR_TYPE_NSSTRING = (
    normalize_uuid_str(CBUUIDCharacteristicUserDescriptionString),  # 0x2901
)


def is_descriptor_nsnumber(value: Any, descriptor_uuid: str) -> TypeGuard[NSNumber]:
    """Check if descriptor value is returned as NSNumber by CoreBluetooth."""
    return descriptor_uuid in _DESCRIPTOR_TYPE_NSNUMBER


def is_descriptor_nsstring(value: Any, descriptor_uuid: str) -> TypeGuard[NSString]:
    """Check if descriptor value is returned as NSString by CoreBluetooth."""
    return descriptor_uuid in _DESCRIPTOR_TYPE_NSSTRING


_CB_MANAGER_STATE_MESSAGE = {
    CBManagerStateUnknown: "Bluetooth in unknown state",
    CBManagerStateResetting: "Bluetooth is resetting",
    CBManagerStateUnsupported: "Bluetooth is unsupported",
    CBManagerStateUnauthorized: "Bluetooth is unauthorized",
    CBManagerStatePoweredOff: "Bluetooth powered off",
    CBManagerStatePoweredOn: "Bluetooth powered on",
}


def cb_manager_state_message(state: CBManagerState) -> str:
    """Get log message for CoreBluetooth manager state."""
    return _CB_MANAGER_STATE_MESSAGE.get(
        state, f"Unknown CBManagerState value: {state}"
    )



import pytest

from bleak import BleakClient, BleakScanner


async def test_adapter_kwarg_deprecated_in_scanner():
    with pytest.deprecated_call(
        match="the 'adapter' keyword argument is deprecated, use the 'bluez' kwarg instead"
    ):
        BleakScanner(adapter="hci0")


async def test_adapter_kwarg_deprecated_in_client():
    with pytest.deprecated_call(
        match="the 'adapter' keyword argument is deprecated, use the 'bluez' kwarg instead"
    ):
        BleakClient("00:11:22:33:44:55", adapter="hci0")



import asyncio
import dataclasses
import sys
from collections.abc import AsyncGenerator

import pytest
import pytest_asyncio
from bumble.att import ATT_Error, ErrorCode
from bumble.device import Connection, Device
from bumble.gatt import Characteristic, CharacteristicValue, Service
from bumble.transport.common import Transport

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakGATTProtocolError
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    create_bumble_peripheral,
    create_hci_transport,
    find_ble_device,
)


@pytest_asyncio.fixture(loop_scope="module", scope="module")
async def hci_transport(
    request: pytest.FixtureRequest,
) -> AsyncGenerator[Transport, None]:
    """Create a bumble HCI Transport."""
    async with create_hci_transport(request) as hci_transport:
        yield hci_transport


@pytest_asyncio.fixture(loop_scope="module", scope="module")
async def bumble_peripheral(hci_transport: Transport) -> Device:
    """Create a BLE peripheral device with bumble."""
    return create_bumble_peripheral(hci_transport)


CHAR_TEST_SERVICE_UUID = "2908f536-3fab-43c9-a7b2-80b6fdaae99b"

READ_CHAR_UUID = "1a1af049-9c23-4b69-b763-e1096674ed18"
WRITE_WITH_RESPONSE_CHAR_UUID = "79a92bad-31b4-4a70-885c-e704ae2c6363"
WRITE_WITHOUT_RESPONSE_CHAR_UUID = "2a07f6ea-6401-45d7-838c-0f459a7edb7f"
NOTIFY_CHAR_UUID = "d4c6dad3-76f1-4034-8871-0a6345be6cfc"
INDICATE_CHAR_UUID = "b1f7c8d4-9e2a-4f6b-8d3a-1234567890ab"


@dataclasses.dataclass
class CharTestPeripheral:
    bleak_client: BleakClient
    bumble_peripheral: Device
    read_characteristic: Characteristic[bytes]
    write_characteristic: Characteristic[bytes]
    write_without_response_characteristic: Characteristic[bytes]
    notify_characteristic: Characteristic[bytes]
    indicate_characteristic: Characteristic[bytes]


@pytest_asyncio.fixture(loop_scope="module", scope="module")
async def char_test_peripheral(
    bumble_peripheral: Device,
) -> AsyncGenerator[CharTestPeripheral, None]:
    read_characteristic = Characteristic[bytes](
        READ_CHAR_UUID,
        Characteristic.Properties.READ,
        Characteristic.Permissions.READABLE,
        b"DATA",
    )
    write_characteristic = Characteristic[bytes](
        WRITE_WITH_RESPONSE_CHAR_UUID,
        Characteristic.Properties.WRITE,
        Characteristic.WRITEABLE,
        b"----",
    )
    write_without_response_characteristic = Characteristic[bytes](
        WRITE_WITHOUT_RESPONSE_CHAR_UUID,
        Characteristic.Properties.WRITE_WITHOUT_RESPONSE,
        Characteristic.WRITEABLE,
        None,
    )
    notify_characteristic = Characteristic[bytes](
        NOTIFY_CHAR_UUID,
        Characteristic.Properties.READ | Characteristic.Properties.NOTIFY,
        Characteristic.Permissions.READABLE,
        b"----",
    )
    indicate_characteristic = Characteristic[bytes](
        INDICATE_CHAR_UUID,
        Characteristic.Properties.READ | Characteristic.Properties.INDICATE,
        Characteristic.Permissions.READABLE,
        b"----",
    )

    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[
            Service(
                CHAR_TEST_SERVICE_UUID,
                [
                    read_characteristic,
                    write_characteristic,
                    write_without_response_characteristic,
                    notify_characteristic,
                    indicate_characteristic,
                ],
            ),
        ],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device, services=[CHAR_TEST_SERVICE_UUID]) as client:
        yield CharTestPeripheral(
            bumble_peripheral=bumble_peripheral,
            bleak_client=client,
            read_characteristic=read_characteristic,
            write_characteristic=write_characteristic,
            write_without_response_characteristic=write_without_response_characteristic,
            notify_characteristic=notify_characteristic,
            indicate_characteristic=indicate_characteristic,
        )


GATT_ERRORS = [
    # These error codes are explicitly called out as possible errors from
    # ATT_WRITE_REQ in the Bluetooth Core Specification. Commented-out codes
    # are not possible to test, e.g. causes BlueZ to disconnect rather than
    # return an error.
    ErrorCode.INVALID_ATTRIBUTE_LENGTH,
    ErrorCode.INSUFFICIENT_AUTHORIZATION,
    # ErrorCode.INSUFFICIENT_AUTHENTICATION,
    # ErrorCode.INSUFFICIENT_ENCRYPTION_KEY_SIZE,
    # ErrorCode.INSUFFICIENT_ENCRYPTION,
    ErrorCode.INVALID_HANDLE,
    #
    # Then these are specifically handled by BlueZ, so assumed to be
    # possible in the wild. INVALID_OFFSET seems to trigger .InvalidArgs
    # which is the same as INVALID_ATTRIBUTE_LENGTH, so not tested here.
    ErrorCode.REQUEST_NOT_SUPPORTED,
    # ErrorCode.INVALID_OFFSET,
    ErrorCode.CCCD_IMPROPERLY_CONFIGURED,
    #
    # And allows custom error codes from higher layers.
    ErrorCode(0xC0),  # Application Error
    ErrorCode(0xE0),  # Common Profile or Service Error Codes
]

GATT_READ_ERRORS = GATT_ERRORS + [ErrorCode.READ_NOT_PERMITTED]
GATT_WRITE_ERRORS = GATT_ERRORS + [ErrorCode.WRITE_NOT_PERMITTED]


@pytest.mark.asyncio(loop_scope="module")
async def test_read_gatt_char(
    char_test_peripheral: CharTestPeripheral,
):
    """Reading a GATT characteristic is possible."""

    # Set data to a known value
    char_test_peripheral.read_characteristic.value = b"DATA"
    data = await char_test_peripheral.bleak_client.read_gatt_char(READ_CHAR_UUID)

    # Verify the data is as expected
    assert data == b"DATA"


@pytest.mark.asyncio(loop_scope="module")
async def test_read_gatt_char_use_cached(char_test_peripheral: CharTestPeripheral):
    """Reading a cached GATT characteristic is possible."""

    # Set data to a known value
    char_test_peripheral.read_characteristic.value = b"ORIGINAL"
    await char_test_peripheral.bleak_client.read_gatt_char(READ_CHAR_UUID)

    # Change the data to a different value
    char_test_peripheral.read_characteristic.value = b"CHANGED"

    data = await char_test_peripheral.bleak_client.read_gatt_char(
        READ_CHAR_UUID, use_cached=True
    )

    # The data has to be the old value since we are using the cached value.
    assert data == b"ORIGINAL"


@pytest.mark.parametrize("gatt_error", GATT_READ_ERRORS)
@pytest.mark.asyncio(loop_scope="module")
async def test_read_gatt_char_error(
    char_test_peripheral: CharTestPeripheral, gatt_error: ErrorCode
):
    """
    Reading a GATT characteristic passes back GATT errors correctly.
    """

    # TODO: should add a test option to allow not skipping if user is available
    # to click through any permission prompts.
    if sys.platform == "darwin" and gatt_error == ErrorCode.INSUFFICIENT_AUTHORIZATION:
        pytest.skip(
            "Apple asks for user permission via a prompt, which cannot be handled automatically."
        )

    def error_on_read(connection: Connection) -> bytes:
        raise ATT_Error(gatt_error)

    char_test_peripheral.read_characteristic.value = CharacteristicValue(
        read=error_on_read
    )

    with pytest.raises(BleakGATTProtocolError) as exc_info:
        await char_test_peripheral.bleak_client.read_gatt_char(READ_CHAR_UUID)

    assert exc_info.value.code == gatt_error


@pytest.mark.asyncio(loop_scope="module")
async def test_write_gatt_char_with_response(char_test_peripheral: CharTestPeripheral):
    """Writing a GATT characteristic is possible."""

    # Set data to a known value
    char_test_peripheral.write_characteristic.value = b"----"

    await char_test_peripheral.bleak_client.write_gatt_char(
        WRITE_WITH_RESPONSE_CHAR_UUID, b"DATA", response=True
    )

    # Verify the new data was written correctly
    assert char_test_peripheral.write_characteristic.value == b"DATA"


@pytest.mark.parametrize("gatt_error", GATT_WRITE_ERRORS)
@pytest.mark.asyncio(loop_scope="module")
async def test_write_gatt_char_with_response_error(
    char_test_peripheral: CharTestPeripheral, gatt_error: ErrorCode
):
    """
    Writing a GATT characteristic propagates error codes correctly.
    """

    # TODO: should add a test option to allow not skipping if user is available
    # to click through any permission prompts.
    if sys.platform == "darwin" and gatt_error == ErrorCode.INSUFFICIENT_AUTHORIZATION:
        pytest.skip(
            "Apple asks for user permission via a prompt, which cannot be handled automatically."
        )

    def error_on_write(connection: Connection, value: bytes):
        raise ATT_Error(gatt_error)

    char_test_peripheral.write_characteristic.value = CharacteristicValue(
        write=error_on_write
    )

    with pytest.raises(BleakGATTProtocolError) as exc_info:
        await char_test_peripheral.bleak_client.write_gatt_char(
            WRITE_WITH_RESPONSE_CHAR_UUID, b"DATA", response=True
        )

    assert exc_info.value.code == gatt_error


@pytest.mark.asyncio(loop_scope="module")
async def test_write_gatt_char_no_response(char_test_peripheral: CharTestPeripheral):
    """Writing a GATT characteristic is possible."""
    peripheral_write_callback_called: asyncio.Future[bytes] = asyncio.Future()

    def peripheral_write_callback(connection: Connection, value: bytes):
        peripheral_write_callback_called.set_result(value)

    char_test_peripheral.write_without_response_characteristic.value = (
        CharacteristicValue(write=peripheral_write_callback)
    )

    await char_test_peripheral.bleak_client.write_gatt_char(
        WRITE_WITHOUT_RESPONSE_CHAR_UUID, b"DATA", response=False
    )
    written_value = await asyncio.wait_for(peripheral_write_callback_called, timeout=1)
    assert written_value == b"DATA"


@pytest.mark.asyncio(loop_scope="module")
async def test_notify_gatt_char(char_test_peripheral: CharTestPeripheral):
    """Writing a GATT characteristic is possible."""

    notified_data: asyncio.Queue[bytes] = asyncio.Queue()

    def notify_callback(characteristic: BleakGATTCharacteristic, data: bytearray):
        assert characteristic.uuid.lower() == NOTIFY_CHAR_UUID
        notified_data.put_nowait(bytes(data))

    await char_test_peripheral.bleak_client.start_notify(
        NOTIFY_CHAR_UUID,
        notify_callback,
    )
    assert notified_data.empty()

    await char_test_peripheral.bumble_peripheral.notify_subscribers(  # type: ignore  # (missing type hints in bumble)
        char_test_peripheral.notify_characteristic,
        b"1234",
    )

    data = await asyncio.wait_for(notified_data.get(), timeout=1)
    assert data == b"1234"

    await char_test_peripheral.bumble_peripheral.notify_subscribers(  # type: ignore  # (missing type hints in bumble)
        char_test_peripheral.notify_characteristic,
        b"2345",
    )

    data = await asyncio.wait_for(notified_data.get(), timeout=1)
    assert data == b"2345"

    await char_test_peripheral.bleak_client.stop_notify(NOTIFY_CHAR_UUID)

    await char_test_peripheral.bumble_peripheral.notify_subscribers(  # type: ignore  # (missing type hints in bumble)
        char_test_peripheral.notify_characteristic,
        b"2345",
    )

    # Verify no notification was received
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(notified_data.get(), timeout=1)


@pytest.mark.asyncio(loop_scope="module")
async def test_indicate_gatt_char(char_test_peripheral: CharTestPeripheral):
    """Ensure indications are delivered and received by the client."""

    virtual_connection = list(
        char_test_peripheral.bumble_peripheral.connections.values()
    )[0]

    indicated_data: asyncio.Queue[bytes] = asyncio.Queue()

    def indicate_callback(characteristic: BleakGATTCharacteristic, data: bytearray):
        assert characteristic.uuid.lower() == INDICATE_CHAR_UUID
        indicated_data.put_nowait(bytes(data))

    await char_test_peripheral.bleak_client.start_notify(
        INDICATE_CHAR_UUID,
        indicate_callback,
    )
    assert indicated_data.empty()

    await char_test_peripheral.bumble_peripheral.indicate_subscriber(  # type: ignore  # (missing type hints in bumble)
        virtual_connection,
        char_test_peripheral.indicate_characteristic,
        b"ind1",
    )

    # dont wait since 'indicate_subscriber' waits for ack from client
    data = indicated_data.get_nowait()
    assert data == b"ind1"

    await char_test_peripheral.bumble_peripheral.indicate_subscriber(  # type: ignore  # (missing type hints in bumble)
        virtual_connection,
        char_test_peripheral.indicate_characteristic,
        b"ind2",
    )

    # dont wait since 'indicate_subscriber' waits for ack from client
    data = indicated_data.get_nowait()
    assert data == b"ind2"

    await char_test_peripheral.bleak_client.stop_notify(INDICATE_CHAR_UUID)

    await char_test_peripheral.bumble_peripheral.indicate_subscriber(  # type: ignore  # (missing type hints in bumble)
        virtual_connection,
        char_test_peripheral.indicate_characteristic,
        b"ind2",
    )

    # Verify no indication was received after stop
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(indicated_data.get(), timeout=1)



import dataclasses
import sys
from collections.abc import AsyncGenerator

import pytest
import pytest_asyncio
from bumble import gatt
from bumble.att import ATT_Error, AttributeValue, ErrorCode
from bumble.core import UUID
from bumble.device import Connection, Device
from bumble.gatt import Characteristic, Descriptor, Service
from bumble.transport.common import Transport

from bleak import BleakClient
from bleak.backends import BleakBackend
from bleak.exc import BleakDBusError, BleakGATTProtocolError
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    create_bumble_peripheral,
    create_hci_transport,
    find_ble_device,
)


@pytest_asyncio.fixture(loop_scope="module", scope="module")
async def hci_transport(
    request: pytest.FixtureRequest,
) -> AsyncGenerator[Transport, None]:
    """Create a bumble HCI Transport."""
    async with create_hci_transport(request) as hci_transport:
        yield hci_transport


@pytest_asyncio.fixture(loop_scope="module", scope="module")
async def bumble_peripheral(hci_transport: Transport) -> Device:
    """Create a BLE peripheral device with bumble."""
    return create_bumble_peripheral(hci_transport)


DESCR_TEST_SERVICE_UUID = "0d15eded-4e68-4718-bedf-736847b68e72"

READABLE_DESCR_CHAR_UUID = "25c614ab-1560-46da-94bc-c146addfc359"
WRITABLE_DESCR_CHAR_UUID = "822afd2f-c2b2-4302-9edb-09850a93b707"

CUSTOM_BINARY_DESCRIPTOR_UUID = UUID("a1b2c3d4-e5f6-7890-1234-56789abcdef0")
# CoreBluetooth specific UUID
L2CAPPPSM_DESCRIPTOR_UUID = UUID("ABDD3056-28FA-441D-A470-55A75A52553A")


@dataclasses.dataclass
class DescrTestPeripheral:
    bleak_client: BleakClient
    bumble_peripheral: Device
    readable_descr: dict[UUID, Descriptor]
    writable_descr: dict[UUID, Descriptor]


STANDARD_DESCRIPTORS: list[tuple[UUID, bytes, bytes]] = [
    (
        gatt.GATT_CHARACTERISTIC_EXTENDED_PROPERTIES_DESCRIPTOR,
        b"\x03\x00",  # Reliable Write + Writable Auxiliaries
        b"\x01\x00",  # Only Reliable Write
    ),
    (
        gatt.GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
        "🚀 Description".encode(),  # Test Description
        "🔄 Updated".encode(),  # Alternative Description
    ),
    (
        gatt.GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
        b"\x01\x00",  # Notifications enabled
        b"\x02\x00",  # Indications enabled
    ),
    (
        gatt.GATT_SERVER_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
        b"\x01\x00",  # Broadcasts enabled
        b"\x00\x00",  # Broadcasts disabled
    ),
    (
        gatt.GATT_CHARACTERISTIC_PRESENTATION_FORMAT_DESCRIPTOR,
        b"\x04\x00\xad\x27\x01\x00\x00",  # uint8, exponent 0, unit temperature celsius
        b"\x06\x00\x72\x27\x01\x00\x00",  # uint16, exponent 0, unit temperature celsius
    ),
    (
        gatt.GATT_CHARACTERISTIC_AGGREGATE_FORMAT_DESCRIPTOR,
        b"\x0a\x00\x0b\x00",  # Handles to two characteristics
        b"\x0c\x00\x0d\x00\x0e\x00",  # Handles to three characteristics
    ),
    (
        gatt.GATT_VALID_RANGE_DESCRIPTOR,
        b"\x00\x00\x64\x00",  # Min: 0, Max: 100
        b"\x0a\x00\x32\x00",  # Min: 10, Max: 50
    ),
    (
        gatt.GATT_EXTERNAL_REPORT_DESCRIPTOR,
        b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f",  # Example UUID
        b"\x0f\x0e\x0d\x0c\x0b\x0a\x09\x08\x07\x06\x05\x04\x03\x02\x01\x00",  # Different UUID
    ),
    (
        gatt.GATT_REPORT_REFERENCE_DESCRIPTOR,
        b"\x01\x01",  # Report ID: 1, Report Type: Input
        b"\x02\x02",  # Report ID: 2, Report Type: Output
    ),
    (
        gatt.GATT_NUMBER_OF_DIGITALS_DESCRIPTOR,
        b"\x08",  # 8 digital values
        b"\x10",  # 16 digital values
    ),
    (
        gatt.GATT_VALUE_TRIGGER_SETTING_DESCRIPTOR,
        b"\x01\x00\x00",  # Condition: None, value: 0
        b"\x02\x32\x00",  # Condition: different, value: 50
    ),
    (
        gatt.GATT_ENVIRONMENTAL_SENSING_CONFIGURATION_DESCRIPTOR,
        b"\x01",  # Trigger logic: OR
        b"\x02",  # Trigger logic: AND
    ),
    (
        gatt.GATT_ENVIRONMENTAL_SENSING_MEASUREMENT_DESCRIPTOR,
        b"\x01\x00\x00",  # Flags: 0x01, Sampling Function: Unspecified, Measurement Period: 0
        b"\x02\x01\x0a",  # Flags: 0x02, Sampling Function: Instantaneous, Measurement Period: 10
    ),
    (
        gatt.GATT_ENVIRONMENTAL_SENSING_TRIGGER_DESCRIPTOR,
        b"\x01\x00\x00",  # Flags: 0x01
        b"\x02\x00\x00",  # Flags: 0x02
    ),
    (
        gatt.GATT_TIME_TRIGGER_DESCRIPTOR,
        b"\x01\x00\x00",  # Condition: None, value: 0
        b"\x02\x3c\x00",  # Condition: different, value: 60
    ),
    (
        gatt.GATT_VALID_RANGE_AND_ACCURACY_DESCRIPTOR,
        b"\x00\x00\x64\x00\x01\x00",  # Min: 0, Max: 100, Accuracy: 1
        b"\x14\x00\x50\x00\x02\x00",  # Min: 20, Max: 80, Accuracy: 2
    ),
    (
        CUSTOM_BINARY_DESCRIPTOR_UUID,
        b"\x01\x02\x03\x04",  # Custom binary data
        b"\xaa\xbb\xcc\xdd",  # Different custom binary data
    ),
    (
        L2CAPPPSM_DESCRIPTOR_UUID,
        b"\x56\x30",  # PSM value 0x3056 in little-endian
        b"\x78\x40",  # PSM value 0x4078 in little-endian
    ),
]


@pytest_asyncio.fixture(loop_scope="module", scope="module")
async def descr_test_peripheral(
    bumble_peripheral: Device,
) -> AsyncGenerator[DescrTestPeripheral, None]:
    readable_descriptors = {
        descr_uuid: Descriptor(descr_uuid, Descriptor.READABLE, descr_data)
        for descr_uuid, descr_data, _ in STANDARD_DESCRIPTORS
    }

    readable_descr_char = Characteristic(
        READABLE_DESCR_CHAR_UUID,
        Characteristic.Properties.READ,
        Characteristic.Permissions.READABLE,
        b"",
        list(readable_descriptors.values()),
    )

    writable_descr = {
        descr_uuid: Descriptor(descr_uuid, Descriptor.WRITEABLE, descr_data)
        for descr_uuid, descr_data, _ in STANDARD_DESCRIPTORS
    }
    writable_descr_char = Characteristic(
        WRITABLE_DESCR_CHAR_UUID,
        Characteristic.Properties.READ,
        Characteristic.Permissions.READABLE,
        b"",
        list(writable_descr.values()),
    )

    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral,
        services=[
            Service(
                DESCR_TEST_SERVICE_UUID,
                [
                    readable_descr_char,
                    writable_descr_char,
                ],
            ),
        ],
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device, services=[DESCR_TEST_SERVICE_UUID]) as client:
        yield DescrTestPeripheral(
            bumble_peripheral=bumble_peripheral,
            bleak_client=client,
            readable_descr=readable_descriptors,
            writable_descr=writable_descr,
        )


GATT_ERRORS = [
    # These error codes are explicitly called out as possible errors from
    # ATT_WRITE_REQ in the Bluetooth Core Specification. Commented-out codes
    # are not possible to test, e.g. causes BlueZ to disconnect rather than
    # return an error.
    ErrorCode.INVALID_ATTRIBUTE_LENGTH,
    ErrorCode.INSUFFICIENT_AUTHORIZATION,
    # ErrorCode.INSUFFICIENT_AUTHENTICATION,
    # ErrorCode.INSUFFICIENT_ENCRYPTION_KEY_SIZE,
    # ErrorCode.INSUFFICIENT_ENCRYPTION,
    ErrorCode.INVALID_HANDLE,
    #
    # Then these are specifically handled by BlueZ, so assumed to be
    # possible in the wild. INVALID_OFFSET seems to trigger .InvalidArgs
    # which is the same as INVALID_ATTRIBUTE_LENGTH, so not tested here.
    ErrorCode.REQUEST_NOT_SUPPORTED,
    # ErrorCode.INVALID_OFFSET,
    ErrorCode.CCCD_IMPROPERLY_CONFIGURED,
    #
    # And allows custom error codes from higher layers.
    ErrorCode(0xC0),  # Application Error
    ErrorCode(0xE0),  # Common Profile or Service Error Codes
]

GATT_READ_ERRORS = GATT_ERRORS + [ErrorCode.READ_NOT_PERMITTED]
GATT_WRITE_ERRORS = GATT_ERRORS + [ErrorCode.WRITE_NOT_PERMITTED]


@pytest.mark.parametrize(
    "descr_uuid,descr_data",
    [
        pytest.param(
            descr_uuid,
            descr_data,
            id=descr_uuid.to_hex_str(),
        )
        for descr_uuid, descr_data, _ in STANDARD_DESCRIPTORS
    ],
)
@pytest.mark.asyncio(loop_scope="module")
async def test_read_gatt_descriptor(
    descr_test_peripheral: DescrTestPeripheral,
    descr_uuid: UUID,
    descr_data: bytes,
):
    """
    Reading a string GATT descriptor is possible.

    On CoreBluetooth some descriptors are returned as NSString or NSNumber instead of NSData.
    This test ensures that all standard descriptors can be read correctly.
    """

    characteristic = descr_test_peripheral.bleak_client.services.get_characteristic(
        READABLE_DESCR_CHAR_UUID
    )
    assert characteristic

    descriptor = characteristic.get_descriptor(descr_uuid.to_hex_str())
    assert descriptor

    # Set data to a known value (other tests may have modified it)
    descr_test_peripheral.readable_descr[descr_uuid].value = descr_data

    data = await descr_test_peripheral.bleak_client.read_gatt_descriptor(descriptor)

    # Verify the data is as expected
    assert data == descr_data


@pytest.mark.asyncio(loop_scope="module")
async def test_read_gatt_descriptor_use_cached(
    descr_test_peripheral: DescrTestPeripheral,
):
    """Reading a cached GATT descriptor is possible."""
    characteristic = descr_test_peripheral.bleak_client.services.get_characteristic(
        READABLE_DESCR_CHAR_UUID
    )
    assert characteristic

    descriptor = characteristic.get_descriptor(
        gatt.GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR.to_hex_str()
    )
    assert descriptor

    bumble_descr = descr_test_peripheral.readable_descr[
        gatt.GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR
    ]

    # Set data to a known value (other tests may have modified it)
    bumble_descr.value = b"Original"

    data = await descr_test_peripheral.bleak_client.read_gatt_descriptor(descriptor)

    # Verify the data is as expected
    assert data == b"Original"

    # Change the value in the peripheral
    bumble_descr.value = b"Changed"

    data = await descr_test_peripheral.bleak_client.read_gatt_descriptor(
        descriptor, use_cached=True
    )

    # The data has to be the original value since we are using the cached value.
    assert data == b"Original"


@pytest.mark.parametrize("gatt_error", GATT_READ_ERRORS)
@pytest.mark.asyncio(loop_scope="module")
async def test_read_gatt_descriptor_error(
    descr_test_peripheral: DescrTestPeripheral, gatt_error: ErrorCode
):
    """
    Reading a string GATT passes error correctly.
    """

    # TODO: should add a test option to allow not skipping if user is available
    # to click through any permission prompts.
    if sys.platform == "darwin" and gatt_error == ErrorCode.INSUFFICIENT_AUTHORIZATION:
        pytest.skip(
            "Apple asks for user permission via a prompt, which cannot be handled automatically."
        )

    characteristic = descr_test_peripheral.bleak_client.services.get_characteristic(
        READABLE_DESCR_CHAR_UUID
    )
    assert characteristic

    descriptor = characteristic.get_descriptor(
        CUSTOM_BINARY_DESCRIPTOR_UUID.to_hex_str()
    )
    assert descriptor

    def error_on_read(connection: Connection) -> bytes:
        raise ATT_Error(gatt_error)

    # Set data to a known value
    descr_test_peripheral.readable_descr[CUSTOM_BINARY_DESCRIPTOR_UUID].value = (
        AttributeValue(read=error_on_read)
    )

    with pytest.raises(BleakGATTProtocolError) as exc_info:
        await descr_test_peripheral.bleak_client.read_gatt_descriptor(descriptor)

    assert exc_info.value.code == gatt_error


@pytest.mark.parametrize(
    "descr_uuid,descr_data_1,descr_data_2",
    [
        pytest.param(
            descr_uuid,
            descr_data_1,
            descr_data_2,
            id=descr_uuid.to_hex_str(),
        )
        for descr_uuid, descr_data_1, descr_data_2 in STANDARD_DESCRIPTORS
    ],
)
@pytest.mark.asyncio(loop_scope="module")
async def test_write_gatt_descriptor(
    descr_test_peripheral: DescrTestPeripheral,
    descr_uuid: UUID,
    descr_data_1: bytes,
    descr_data_2: bytes,
):
    """Writing a GATT descriptor is possible."""

    characteristic = descr_test_peripheral.bleak_client.services.get_characteristic(
        WRITABLE_DESCR_CHAR_UUID
    )
    assert characteristic

    descriptor = characteristic.get_descriptor(descr_uuid.to_hex_str())
    assert descriptor

    # Set data to a known value
    descr_test_peripheral.writable_descr[descr_uuid].value = descr_data_1

    if descr_uuid == gatt.GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR:
        # writing CCCD should raise ValueError
        with pytest.raises(ValueError, match="Cannot write to CCCD"):
            await descr_test_peripheral.bleak_client.write_gatt_descriptor(
                descriptor, descr_data_2
            )
    else:
        await descr_test_peripheral.bleak_client.write_gatt_descriptor(
            descriptor, descr_data_2
        )

        # Verify the data is as expected
        assert descr_test_peripheral.writable_descr[descr_uuid].value == descr_data_2  # type: ignore  # (missing type hints in bumble)


@pytest.mark.parametrize("gatt_error", GATT_WRITE_ERRORS)
@pytest.mark.asyncio(loop_scope="module")
async def test_write_gatt_descriptor_error(
    descr_test_peripheral: DescrTestPeripheral, gatt_error: ErrorCode
):
    """
    Writing a string GATT passes error correctly.
    """

    # TODO: should add a test option to allow not skipping if user is available
    # to click through any permission prompts.
    if sys.platform == "darwin" and gatt_error == ErrorCode.INSUFFICIENT_AUTHORIZATION:
        pytest.skip(
            "Apple asks for user permission via a prompt, which cannot be handled automatically."
        )

    characteristic = descr_test_peripheral.bleak_client.services.get_characteristic(
        WRITABLE_DESCR_CHAR_UUID
    )
    assert characteristic

    descriptor = characteristic.get_descriptor(
        CUSTOM_BINARY_DESCRIPTOR_UUID.to_hex_str()
    )
    assert descriptor

    def error_on_write(connection: Connection, data: bytes) -> None:
        raise ATT_Error(gatt_error)

    # Set data to a known value
    descr_test_peripheral.writable_descr[CUSTOM_BINARY_DESCRIPTOR_UUID].value = (
        AttributeValue(write=error_on_write)
    )

    with pytest.raises(BleakGATTProtocolError) as exc_info:
        await descr_test_peripheral.bleak_client.write_gatt_descriptor(descriptor, b"")

    assert exc_info.value.code == gatt_error


@pytest.mark.asyncio(loop_scope="module")
async def test_notify_gatt_char_error(descr_test_peripheral: DescrTestPeripheral):
    """
    Failing to enable notifications passes error correctly.

    The CCCD descriptor can't be written directly, so we have to use the
    start_notify() method to trigger the write.
    """

    client = descr_test_peripheral.bleak_client

    # BlueZ will just disconnect the device instead of passing along any error
    # when enabling notifications fails. It seems the best we can do is just
    # check a BlueZ quirk where if a characteristic doesn't have the notify
    # property, we get a D-Bus error without doing any I/O.

    if client.backend_id != BleakBackend.BLUEZ_DBUS:
        # TODO: See how other backends respond to similar errors.
        pytest.skip("This test relies on a BlueZ-specific quirk.")

    characteristic = client.services.get_characteristic(WRITABLE_DESCR_CHAR_UUID)
    assert characteristic

    def error_on_write(connection: Connection, data: bytes) -> None:
        raise ATT_Error(ErrorCode.INSUFFICIENT_ENCRYPTION)

    # Set data to a known value
    descr_test_peripheral.writable_descr[
        gatt.GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR
    ].value = AttributeValue(write=error_on_write)

    with pytest.raises(BleakDBusError) as exc_info:
        await client.start_notify(characteristic, lambda c, d: None)

    assert exc_info.value.dbus_error == "org.bluez.Error.NotSupported"



import asyncio

import pytest
from bumble.att import Attribute, AttributeValue
from bumble.device import Connection, Device
from bumble.gatt import (
    GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
    Characteristic,
    Descriptor,
    Service,
)

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from tests.integration.conftest import (
    configure_and_power_on_bumble_peripheral,
    find_ble_device,
)

TEST_SERVICE_UUID = "9d513f40-5c89-42dc-9688-2cfa30f2d9e7"
TEST_CHARACTERISTIC_UUID = "e809cb2f-34e3-42a1-ba92-22db2495cd6a"


@pytest.mark.parametrize("use_start_notify", [True, False])
async def test_notification_sent_before_write_response(
    use_start_notify: bool,
    bumble_peripheral: Device,
) -> None:
    """
    Regression test for <https://github.com/hbldh/bleak/issues/1885>.
    """

    notifications_enabled = False

    def on_cccd_read(connection: Connection) -> bytes:
        return b"\x01\x00" if notifications_enabled else b"\x00\x00"

    async def on_cccd_write(connection: Connection, value: bytes) -> None:
        nonlocal notifications_enabled
        notifications_enabled = value == b"\x01\x00"

        # This is simulating an unusual peripheral that sends a notification
        # immediately upon receiving the CCCD write, before sending the write
        # response.

        # TODO: Type hints in bumble need to be fixed to be able to remove the pyright ignore
        await bumble_peripheral.notify_subscribers(  # pyright: ignore[reportUnknownMemberType]
            test_characteristic, b"test", force=True
        )

    cccd_value = AttributeValue[bytes](on_cccd_read, on_cccd_write)

    test_characteristic = Characteristic[bytes](
        TEST_CHARACTERISTIC_UUID,
        Characteristic.Properties.NOTIFY,
        Attribute.Permissions(0),
        descriptors=[
            Descriptor(
                GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
                Attribute.Permissions.WRITEABLE | Attribute.Permissions.READABLE,
                cccd_value,
            )
        ],
    )

    await configure_and_power_on_bumble_peripheral(
        bumble_peripheral, services=[Service(TEST_SERVICE_UUID, [test_characteristic])]
    )

    device = await find_ble_device(bumble_peripheral)

    async with BleakClient(device, services=[TEST_SERVICE_UUID]) as client:
        notification_queue: asyncio.Queue[bytes] = asyncio.Queue()

        def on_notification(_: BleakGATTCharacteristic, data: bytearray) -> None:
            notification_queue.put_nowait(bytes(data))

        await client.start_notify(
            TEST_CHARACTERISTIC_UUID,
            on_notification,
            bluez={"use_start_notify": use_start_notify},
        )

        # In BlueZ, the notification is not received when using "AcquireNotify"
        # causing this to timeout.

        if use_start_notify:
            data = await asyncio.wait_for(notification_queue.get(), timeout=3)

            assert data == b"test"
        else:
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(notification_queue.get(), timeout=3)

        # Normally we don't bother with calling stop_notify() but we are doing
        # it here just to get code coverage since this is the only test where
        # we are setting use_start_notify.
        await client.stop_notify(TEST_CHARACTERISTIC_UUID)



#!/usr/bin/env python

"""Tests for `bleak` package."""

import platform

import pytest

from bleak import BleakClient, BleakScanner
from bleak.backends import BleakBackend
from bleak.backends.client import get_platform_client_backend_type
from bleak.backends.scanner import get_platform_scanner_backend_type


def test_platform_detection():
    """Test by importing the client and assert correct client by OS."""

    client_backend_type, client_backend_id = get_platform_client_backend_type()
    scanner_backend_type, scanner_backend_id = get_platform_scanner_backend_type()

    if platform.system() == "Linux":
        assert client_backend_type.__name__ == "BleakClientBlueZDBus"
        assert scanner_backend_type.__name__ == "BleakScannerBlueZDBus"
        assert client_backend_id == BleakBackend.BLUEZ_DBUS
        assert scanner_backend_id == BleakBackend.BLUEZ_DBUS
    elif platform.system() == "Windows":
        assert client_backend_type.__name__ == "BleakClientWinRT"
        assert scanner_backend_type.__name__ == "BleakScannerWinRT"
        assert client_backend_id == BleakBackend.WIN_RT
        assert scanner_backend_id == BleakBackend.WIN_RT
    elif platform.system() == "Darwin":
        assert client_backend_type.__name__ == "BleakClientCoreBluetooth"
        assert scanner_backend_type.__name__ == "BleakScannerCoreBluetooth"
        assert client_backend_id == BleakBackend.CORE_BLUETOOTH
        assert scanner_backend_id == BleakBackend.CORE_BLUETOOTH


@pytest.mark.asyncio
async def test_backend_id():
    """Test that backend IDs are correct."""
    client = BleakClient("")
    scanner = BleakScanner()

    assert client.backend_id == scanner.backend_id

    if platform.system() == "Linux":
        assert client.backend_id == BleakBackend.BLUEZ_DBUS
        assert scanner.backend_id == BleakBackend.BLUEZ_DBUS
    elif platform.system() == "Windows":
        assert client.backend_id == BleakBackend.WIN_RT
        assert scanner.backend_id == BleakBackend.WIN_RT
    elif platform.system() == "Darwin":
        assert client.backend_id == BleakBackend.CORE_BLUETOOTH
        assert scanner.backend_id == BleakBackend.CORE_BLUETOOTH



"""
TI CC2650 SensorTag
-------------------

An example connecting to a TI CC2650 SensorTag.

Created on 2018-01-10 by hbldh <henrik.blidh@nedomkull.com>

"""

import asyncio
import platform
import sys

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.uuids import normalize_uuid_16, uuid16_dict

ADDRESS = (
    "24:71:89:cc:09:05"
    if platform.system() != "Darwin"
    else "B9EA5233-37EF-4DD6-87A8-2A875E821C46"
)

ALL_SENSORTAG_CHARACTERISTIC_UUIDS = """
00002a00-0000-1000-8000-00805f9b34fb
00002a01-0000-1000-8000-00805f9b34fb
00002a04-0000-1000-8000-00805f9b34fb
00002a23-0000-1000-8000-00805f9b34fb
00002a24-0000-1000-8000-00805f9b34fb
00002a25-0000-1000-8000-00805f9b34fb
00002a26-0000-1000-8000-00805f9b34fb
00002a27-0000-1000-8000-00805f9b34fb
00002a28-0000-1000-8000-00805f9b34fb
00002a29-0000-1000-8000-00805f9b34fb
00002a2a-0000-1000-8000-00805f9b34fb
00002a50-0000-1000-8000-00805f9b34fb
00002a19-0000-1000-8000-00805f9b34fb
f000aa01-0451-4000-b000-000000000000
f000aa02-0451-4000-b000-000000000000
f000aa03-0451-4000-b000-000000000000
f000aa21-0451-4000-b000-000000000000
f000aa22-0451-4000-b000-000000000000
f000aa23-0451-4000-b000-000000000000
f000aa41-0451-4000-b000-000000000000
f000aa42-0451-4000-b000-000000000000
f000aa44-0451-4000-b000-000000000000
f000aa81-0451-4000-b000-000000000000
f000aa82-0451-4000-b000-000000000000
f000aa83-0451-4000-b000-000000000000
f000aa71-0451-4000-b000-000000000000
f000aa72-0451-4000-b000-000000000000
f000aa73-0451-4000-b000-000000000000
0000ffe1-0000-1000-8000-00805f9b34fb
f000aa65-0451-4000-b000-000000000000
f000aa66-0451-4000-b000-000000000000
f000ac01-0451-4000-b000-000000000000
f000ac02-0451-4000-b000-000000000000
f000ac03-0451-4000-b000-000000000000
f000ccc1-0451-4000-b000-000000000000
f000ccc2-0451-4000-b000-000000000000
f000ccc3-0451-4000-b000-000000000000
f000ffc1-0451-4000-b000-000000000000
f000ffc2-0451-4000-b000-000000000000
f000ffc3-0451-4000-b000-000000000000
f000ffc4-0451-4000-b000-000000000000
"""

uuid16_lookup = {v: normalize_uuid_16(k) for k, v in uuid16_dict.items()}

SYSTEM_ID_UUID = uuid16_lookup["System ID"]
MODEL_NBR_UUID = uuid16_lookup["Model Number String"]
DEVICE_NAME_UUID = uuid16_lookup["Device Name"]
FIRMWARE_REV_UUID = uuid16_lookup["Firmware Revision String"]
HARDWARE_REV_UUID = uuid16_lookup["Hardware Revision String"]
SOFTWARE_REV_UUID = uuid16_lookup["Software Revision String"]
MANUFACTURER_NAME_UUID = uuid16_lookup["Manufacturer Name String"]
BATTERY_LEVEL_UUID = uuid16_lookup["Battery Level"]
KEY_PRESS_UUID = normalize_uuid_16(0xFFE1)
# I/O test points on SensorTag.
IO_DATA_CHAR_UUID = "f000aa65-0451-4000-b000-000000000000"
IO_CONFIG_CHAR_UUID = "f000aa66-0451-4000-b000-000000000000"


async def main(address: str):
    async with BleakClient(address, winrt={"use_cached_services": True}) as client:
        print(f"Connected: {client.is_connected}")

        system_id = await client.read_gatt_char(SYSTEM_ID_UUID)
        print(f"System ID: {system_id[::-1].hex(':')}")

        model_number = await client.read_gatt_char(MODEL_NBR_UUID)
        print(f"Model Number: {model_number.decode()}")

        try:
            device_name = await client.read_gatt_char(DEVICE_NAME_UUID)
            print(f"Device Name: {device_name.decode()}")
        except Exception:
            pass

        manufacturer_name = await client.read_gatt_char(MANUFACTURER_NAME_UUID)
        print(f"Manufacturer Name: {manufacturer_name.decode()}")

        firmware_revision = await client.read_gatt_char(FIRMWARE_REV_UUID)
        print(f"Firmware Revision: {firmware_revision.decode()}")

        hardware_revision = await client.read_gatt_char(HARDWARE_REV_UUID)
        print(f"Hardware Revision: {hardware_revision.decode()}")

        software_revision = await client.read_gatt_char(SOFTWARE_REV_UUID)
        print(f"Software Revision: {software_revision.decode()}")

        battery_level = await client.read_gatt_char(BATTERY_LEVEL_UUID)
        print(f"Battery Level: {battery_level[0]}%")

        async def notification_handler(
            characteristic: BleakGATTCharacteristic, data: bytearray
        ):
            print(f"{characteristic.description}: {data}")

        # Turn on the red light on the Sensor Tag by writing to I/O Data and I/O Config.
        write_value = bytearray([0x01])
        value = await client.read_gatt_char(IO_DATA_CHAR_UUID)
        print(f"I/O Data Pre-Write Value: {value}")

        await client.write_gatt_char(IO_DATA_CHAR_UUID, write_value, response=True)

        value = await client.read_gatt_char(IO_DATA_CHAR_UUID)
        print(f"I/O Data Post-Write Value: {value}")
        assert value == write_value

        write_value = bytearray([0x01])
        value = await client.read_gatt_char(IO_CONFIG_CHAR_UUID)
        print(f"I/O Config Pre-Write Value: {value}")

        await client.write_gatt_char(IO_CONFIG_CHAR_UUID, write_value, response=True)

        value = await client.read_gatt_char(IO_CONFIG_CHAR_UUID)
        print(f"I/O Config Post-Write Value: {value}")
        assert value == write_value

        # Try notifications with key presses.

        await client.start_notify(KEY_PRESS_UUID, notification_handler)
        await asyncio.sleep(5.0)
        await client.stop_notify(KEY_PRESS_UUID)


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1] if len(sys.argv) == 2 else ADDRESS))



"""
Disconnect callback
-------------------

An example showing how the `set_disconnected_callback` can be used on BlueZ backend.

Updated on 2019-09-07 by hbldh <henrik.blidh@gmail.com>

"""

import argparse
import asyncio
import logging
from typing import Optional

from bleak import BleakClient, BleakScanner

logger = logging.getLogger(__name__)


class Args(argparse.Namespace):
    name: Optional[str]
    address: Optional[str]
    macos_use_bdaddr: bool
    debug: bool


async def main(args: Args):
    logger.info("scanning...")

    if args.address:
        device = await BleakScanner.find_device_by_address(
            args.address, cb={"use_bdaddr": args.macos_use_bdaddr}
        )
        if device is None:
            logger.error("could not find device with address '%s'", args.address)
            return
    elif args.name:
        device = await BleakScanner.find_device_by_name(
            args.name, cb={"use_bdaddr": args.macos_use_bdaddr}
        )
        if device is None:
            logger.error("could not find device with name '%s'", args.name)
            return
    else:
        raise ValueError("Either --name or --address must be provided")

    disconnected_event = asyncio.Event()

    def disconnected_callback(client: BleakClient):
        logger.info("Disconnected callback called!")
        disconnected_event.set()

    async with BleakClient(
        device, disconnected_callback=disconnected_callback
    ) as client:
        logger.info("Sleeping until device disconnects...")
        await disconnected_event.wait()
        logger.info("Connected: %r", client.is_connected)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    device_group = parser.add_mutually_exclusive_group(required=True)

    device_group.add_argument(
        "--name",
        metavar="<name>",
        help="the name of the bluetooth device to connect to",
    )
    device_group.add_argument(
        "--address",
        metavar="<address>",
        help="the address of the bluetooth device to connect to",
    )

    parser.add_argument(
        "--macos-use-bdaddr",
        action="store_true",
        help="when true use Bluetooth address instead of UUID on macOS",
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="sets the log level to debug",
    )

    args = parser.parse_args(namespace=Args())

    log_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(
        level=log_level,
        format="%(asctime)-15s %(name)-8s %(levelname)s: %(message)s",
    )

    asyncio.run(main(args))



"""
Python version compatibility imports.

These will be removed when support for older Python versions is dropped.
"""

import sys

if sys.version_info < (3, 11):
    from async_timeout import timeout as timeout
    from typing_extensions import Never as Never
    from typing_extensions import Self as Self
    from typing_extensions import TypeVarTuple as TypeVarTuple
    from typing_extensions import Unpack as Unpack
    from typing_extensions import assert_never as assert_never
else:
    from asyncio import timeout as timeout  # noqa: F401
    from typing import Never as Never  # noqa: F401
    from typing import Self as Self  # noqa: F401
    from typing import TypeVarTuple as TypeVarTuple  # noqa: F401
    from typing import Unpack as Unpack  # noqa: F401
    from typing import assert_never as assert_never  # noqa: F401

if sys.version_info < (3, 12):
    from typing_extensions import override as override
else:
    from typing import override as override  # noqa: F401



"""
Scan/Discovery Async Iterator
--------------

Example showing how to scan for BLE devices using async iterator instead of callback function

Created on 2023-07-07 by bojanpotocnik <info@bojanpotocnik.com>

"""

import asyncio

from bleak import BleakScanner


async def main():
    async with BleakScanner() as scanner:
        print("Scanning...")

        n = 5
        print(f"\n{n} advertisement packets:")
        async for bd, ad in scanner.advertisement_data():
            print(f" {n}. {bd!r} with {ad!r}")
            n -= 1
            if n == 0:
                break

        n = 10
        print(f"\nFind device with name longer than {n} characters...")
        async for bd, ad in scanner.advertisement_data():
            found = len(bd.name or "") > n or len(ad.local_name or "") > n
            print(f" Found{' it' if found else ''} {bd!r} with {ad!r}")
            if found:
                break


if __name__ == "__main__":
    asyncio.run(main())



# Created on June, 25 2019 by kevincar <kevincarrolldavis@gmail.com>
"""
CentralManagerDelegate will implement the CBCentralManagerDelegate protocol to
manage CoreBluetooth services and resources on the Central End
"""
from __future__ import annotations

import sys
import weakref
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

import asyncio
import logging
from collections.abc import Callable
from typing import Any, Optional, TypedDict, cast

import objc
from CoreBluetooth import (
    CBUUID,
    CBCentralManager,
    CBManagerAuthorizationDenied,
    CBManagerAuthorizationRestricted,
    CBManagerStatePoweredOff,
    CBManagerStatePoweredOn,
    CBManagerStateResetting,
    CBManagerStateUnauthorized,
    CBManagerStateUnsupported,
    CBPeripheral,
)
from Foundation import (
    NSUUID,
    NSArray,
    NSData,
    NSDictionary,
    NSError,
    NSNumber,
    NSObject,
    NSString,
)
from libdispatch import DISPATCH_QUEUE_SERIAL, dispatch_queue_create

from bleak._compat import Self
from bleak._compat import timeout as async_timeout
from bleak.backends._utils import external_thread_callback, try_call_soon_threadsafe
from bleak.backends.corebluetooth.utils import cb_manager_state_message
from bleak.exc import (
    BleakBluetoothNotAvailableError,
    BleakBluetoothNotAvailableReason,
    BleakError,
)

logger = logging.getLogger(__name__)
CBCentralManagerDelegate = objc.protocolNamed("CBCentralManagerDelegate")


DisconnectCallback = Callable[[], None]


class CBAdvertisementData(TypedDict, total=False):
    kCBAdvDataLocalName: NSString
    kCBAdvDataManufacturerData: NSData
    kCBAdvDataServiceData: dict[CBUUID, NSData]
    kCBAdvDataServiceUUIDs: NSArray[CBUUID]
    kCBAdvertisementDataOverflowServiceUUIDsKey: NSArray[CBUUID]
    kCBAdvDataTxPowerLevel: NSNumber
    kCBAdvertisementDataIsConnectable: NSNumber
    kCBAdvDataOverflowServiceUUIDs: NSArray[CBUUID]


class ObjcCentralManagerDelegate(NSObject, protocols=[CBCentralManagerDelegate]):
    """
    CoreBluetooth central manager delegate for bridging callbacks to asyncio.
    """

    def initWithPyDelegate_(
        self, py_delegate: weakref.ReferenceType["CentralManagerDelegate"]
    ) -> Optional[Self]:
        """macOS init function for NSObject"""
        self = objc.super(ObjcCentralManagerDelegate, self).init()  # type: ignore[assignment]

        if self is None:
            return None  # pragma: no cover

        self.py_delegate = py_delegate

        return self

    # Protocol Functions

    @external_thread_callback
    def centralManagerDidUpdateState_(self, centralManager: CBCentralManager) -> None:
        logger.debug("centralManagerDidUpdateState_")
        logger.debug(cb_manager_state_message(centralManager.state()))

        if (py_delegate := self.py_delegate()) is None:
            return

        try_call_soon_threadsafe(
            py_delegate.event_loop,
            py_delegate.did_update_state_event.set,
        )

    @external_thread_callback
    def centralManager_didDiscoverPeripheral_advertisementData_RSSI_(
        self,
        central: CBCentralManager,
        peripheral: CBPeripheral,
        advertisementData: NSDictionary[str, Any],
        RSSI: NSNumber,
    ) -> None:
        logger.debug("centralManager_didDiscoverPeripheral_advertisementData_RSSI_")

        if (py_delegate := self.py_delegate()) is None:
            return

        try_call_soon_threadsafe(
            py_delegate.event_loop,
            py_delegate.did_discover_peripheral,
            central,
            peripheral,
            advertisementData,
            RSSI,
        )

    @external_thread_callback
    def centralManager_didConnectPeripheral_(
        self, central: CBCentralManager, peripheral: CBPeripheral
    ) -> None:
        logger.debug("centralManager_didConnectPeripheral_")

        if (py_delegate := self.py_delegate()) is None:
            return

        try_call_soon_threadsafe(
            py_delegate.event_loop,
            py_delegate.did_connect_peripheral,
            central,
            peripheral,
        )

    @external_thread_callback
    def centralManager_didFailToConnectPeripheral_error_(
        self,
        centralManager: CBCentralManager,
        peripheral: CBPeripheral,
        error: Optional[NSError],
    ) -> None:
        logger.debug("centralManager_didFailToConnectPeripheral_error_")

        if (py_delegate := self.py_delegate()) is None:
            return

        try_call_soon_threadsafe(
            py_delegate.event_loop,
            py_delegate.did_fail_to_connect_peripheral,
            centralManager,
            peripheral,
            error,
        )

    @external_thread_callback
    def centralManager_didDisconnectPeripheral_error_(
        self,
        central: CBCentralManager,
        peripheral: CBPeripheral,
        error: Optional[NSError],
    ) -> None:
        logger.debug("centralManager_didDisconnectPeripheral_error_")

        if (py_delegate := self.py_delegate()) is None:
            return

        try_call_soon_threadsafe(
            py_delegate.event_loop,
            py_delegate.did_disconnect_peripheral,
            central,
            peripheral,
            error,
        )


class CentralManagerDelegate:
    """
    macOS conforming python class for managing the CentralManger for BLE

    Before this object can be used, the :method:`wait_until_ready` method has to
    be called. This can take a while, when the OS asks the user for permissions.
    """

    def __init__(self) -> None:
        """macOS init function for NSObject"""
        delegate = ObjcCentralManagerDelegate.alloc().initWithPyDelegate_(
            weakref.ref(self)
        )
        assert delegate is not None
        self.objc_delegate = delegate

        self.event_loop = asyncio.get_running_loop()
        self._connect_futures: dict[NSUUID, asyncio.Future[bool]] = {}

        self.callbacks: dict[
            int,
            Callable[[CBPeripheral, CBAdvertisementData, NSNumber], None],
        ] = {}
        self._disconnect_callbacks: dict[NSUUID, DisconnectCallback] = {}
        self._disconnect_futures: dict[NSUUID, asyncio.Future[None]] = {}

        self.did_update_state_event = asyncio.Event()
        self.central_manager = CBCentralManager.alloc().initWithDelegate_queue_(
            self.objc_delegate,
            dispatch_queue_create(b"bleak.corebluetooth", DISPATCH_QUEUE_SERIAL),
        )

    # User defined functions
    @objc.python_method
    async def wait_until_ready(self):
        # According to CoreBluetooth docs, it is not valid to call CBCentral
        # methods until the centralManagerDidUpdateState_() delegate method
        # is called and the current state is CBManagerStatePoweredOn.
        # Wait until the callback occurs. This normally should not take too long,
        # but if the app currently has no permission to access the Bluetooth peripheral,
        # there is automatically a dialog shown by the OS. The user has to accept or deny
        # the Bluetooth access. This may take infinite time until the user clicks something.
        await self.did_update_state_event.wait()

        state = self.central_manager.state()
        if state == CBManagerStateUnsupported:
            raise BleakBluetoothNotAvailableError(
                "Bluetooth is unsupported",
                BleakBluetoothNotAvailableReason.NO_BLUETOOTH,
            )
        elif state == CBManagerStateUnauthorized:
            authorization = self.central_manager.authorization()
            if authorization == CBManagerAuthorizationDenied:
                raise BleakBluetoothNotAvailableError(
                    "Bluetooth access is denied by the user for the current application. Check macOS privacy settings.",
                    BleakBluetoothNotAvailableReason.DENIED_BY_USER,
                )
            elif authorization == CBManagerAuthorizationRestricted:
                raise BleakBluetoothNotAvailableError(
                    "Bluetooth access is restricted for the current application, e.g. by parental controls. Ask the admin to remove this restriction.",
                    BleakBluetoothNotAvailableReason.DENIED_BY_SYSTEM,
                )
            else:
                raise BleakBluetoothNotAvailableError(
                    "Bluetooth is not authorized for an unknown reason. Check macOS privacy settings.",
                    BleakBluetoothNotAvailableReason.DENIED_BY_UNKNOWN,
                )
        elif state == CBManagerStatePoweredOff:
            raise BleakBluetoothNotAvailableError(
                "Bluetooth device is turned off",
                BleakBluetoothNotAvailableReason.POWERED_OFF,
            )
        elif state == CBManagerStateResetting:
            raise BleakBluetoothNotAvailableError(
                "Connection to the Bluetooth system service was lost. Currently trying to reconnect...",
                BleakBluetoothNotAvailableReason.UNKNOWN,
            )
        elif state != CBManagerStatePoweredOn:
            raise BleakBluetoothNotAvailableError(
                "Bluetooth state is unknown",
                BleakBluetoothNotAvailableReason.UNKNOWN,
            )

    async def start_scan(self, service_uuids: Optional[list[str]]) -> None:
        _service_uuids = (
            NSArray[CBUUID]
            .alloc()
            .initWithArray_(list(map(CBUUID.UUIDWithString_, service_uuids)))
            if service_uuids
            else None
        )

        self.central_manager.scanForPeripheralsWithServices_options_(
            _service_uuids, None
        )

    async def stop_scan(self) -> None:
        self.central_manager.stopScan()

    async def connect(
        self,
        peripheral: CBPeripheral,
        disconnect_callback: DisconnectCallback,
        timeout: float = 10.0,
    ) -> None:
        try:
            self._disconnect_callbacks[peripheral.identifier()] = disconnect_callback
            future = self.event_loop.create_future()

            self._connect_futures[peripheral.identifier()] = future
            try:
                self.central_manager.connectPeripheral_options_(peripheral, None)
                async with async_timeout(timeout):
                    await future
            finally:
                del self._connect_futures[peripheral.identifier()]

        except asyncio.TimeoutError:
            logger.debug(f"Connection timed out after {timeout} seconds.")
            del self._disconnect_callbacks[peripheral.identifier()]
            future = self.event_loop.create_future()

            self._disconnect_futures[peripheral.identifier()] = future
            try:
                self.central_manager.cancelPeripheralConnection_(peripheral)
                await future
            finally:
                del self._disconnect_futures[peripheral.identifier()]

            raise

    async def disconnect(self, peripheral: CBPeripheral) -> None:
        future = self.event_loop.create_future()

        self._disconnect_futures[peripheral.identifier()] = future
        try:
            self.central_manager.cancelPeripheralConnection_(peripheral)
            await future
        finally:
            del self._disconnect_futures[peripheral.identifier()]

    # Protocol Functions

    def did_discover_peripheral(
        self,
        central: CBCentralManager,
        peripheral: CBPeripheral,
        advertisementData: NSDictionary[str, Any],
        RSSI: NSNumber,
    ) -> None:
        # Note: this function might be called several times for same device.
        # This can happen for instance when an active scan is done, and the
        # second call with contain the data from the BLE scan response.
        # Example a first time with the following keys in advertisementData:
        # ['kCBAdvDataLocalName', 'kCBAdvDataIsConnectable', 'kCBAdvDataChannel']
        # ... and later a second time with other keys (and values) such as:
        # ['kCBAdvDataServiceUUIDs', 'kCBAdvDataIsConnectable', 'kCBAdvDataChannel']
        #
        # i.e it is best not to trust advertisementData for later use and data
        # from it should be copied.
        #
        # This behaviour could be affected by the
        # CBCentralManagerScanOptionAllowDuplicatesKey global setting.

        uuid_string = peripheral.identifier().UUIDString()

        for callback in self.callbacks.values():
            callback(peripheral, cast(CBAdvertisementData, advertisementData), RSSI)

        logger.debug(
            "Discovered device %s: %s @ RSSI: %d (kCBAdvData %r) and Central: %r",
            uuid_string,
            peripheral.name(),
            RSSI,
            advertisementData.keys(),
            central,
        )

    def did_connect_peripheral(
        self, central: CBCentralManager, peripheral: CBPeripheral
    ) -> None:
        future = self._connect_futures.get(peripheral.identifier(), None)
        if future is not None:
            future.set_result(True)

    def did_fail_to_connect_peripheral(
        self,
        centralManager: CBCentralManager,
        peripheral: CBPeripheral,
        error: Optional[NSError],
    ) -> None:
        future = self._connect_futures.get(peripheral.identifier(), None)
        if future is not None:
            if error is not None:
                future.set_exception(BleakError(f"failed to connect: {error}"))
            else:
                future.set_result(False)

    def did_disconnect_peripheral(
        self,
        central: CBCentralManager,
        peripheral: CBPeripheral,
        error: Optional[NSError],
    ) -> None:
        logger.debug("Peripheral Device disconnected!")

        future = self._disconnect_futures.get(peripheral.identifier(), None)
        if future is not None:
            if error is not None:
                future.set_exception(BleakError(f"disconnect failed: {error}"))
            else:
                future.set_result(None)

        callback = self._disconnect_callbacks.pop(peripheral.identifier(), None)

        if callback is not None:
            callback()



# Created on 2018-04-23 by hbldh <henrik.blidh@nedomkull.com>
"""
Wrapper class for Bluetooth LE servers returned from calling
:py:meth:`bleak.discover`.
"""


from typing import Any, Optional
from warnings import warn


class BLEDevice:
    """
    A simple wrapper class representing a BLE server detected during scanning.
    """

    __slots__ = ("address", "name", "details")

    def __init__(self, address: str, name: Optional[str], details: Any, **kwargs: Any):
        #: The Bluetooth address of the device on this machine (UUID on macOS).
        self.address = address
        #: The operating system name of the device (not necessarily the local name
        #: from the advertising data), suitable for display to the user.
        self.name = name
        #: The OS native details required for connecting to the device.
        self.details = details

        if kwargs:
            warn(
                "Passing additional arguments for BLEDevice is deprecated and has no effect.",
                DeprecationWarning,
                stacklevel=2,
            )

    def __str__(self):
        return f"{self.address}: {self.name}"

    def __repr__(self):
        return f"BLEDevice({self.address}, {self.name})"
